# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import abc
import logging
import time
from typing import TYPE_CHECKING, Any, Optional

from azure.ai.ml.exceptions import ErrorCategory, ErrorTarget, ValidationException

from ..._constants import DEFAULT_REFRESH_OFFSET, DEFAULT_TOKEN_REFRESH_RETRY_DELAY

if TYPE_CHECKING:
    from azure.core.credentials import AccessToken

_LOGGER = logging.getLogger(__name__)


class GetTokenMixin(abc.ABC):
    def __init__(self, *args: "Any", **kwargs: "Any") -> None:
        self._last_request_time = 0

        # https://github.com/python/mypy/issues/5887
        super(GetTokenMixin, self).__init__(*args, **kwargs)

    @abc.abstractmethod
    # pylint: disable-next=docstring-missing-param
    async def _acquire_token_silently(self, *scopes: str, **kwargs: "Any") -> "Optional[AccessToken]":
        """Attempt to acquire an access token from a cache or by redeeming a refresh token."""

    @abc.abstractmethod
    # pylint: disable-next=docstring-missing-param
    async def _request_token(self, *scopes: str, **kwargs: "Any") -> "AccessToken":
        """Request an access token from the STS."""

    def _should_refresh(self, token: "AccessToken") -> bool:
        now = int(time.time())
        if token.expires_on - now > DEFAULT_REFRESH_OFFSET:
            return False
        if now - self._last_request_time < DEFAULT_TOKEN_REFRESH_RETRY_DELAY:
            return False
        return True

    async def get_token(self, *scopes: str, **kwargs: "Any") -> "AccessToken":
        """Request an access token for `scopes`.

        This method is called automatically by Azure SDK clients.

        :param scopes: Scopes to request access for
        :type: str
        :raises CredentialUnavailableError: the credential is unable to attempt authentication because it lacks
            required data, state, or platform support
        :raises ~azure.core.exceptions.ClientAuthenticationError: authentication failed. The error's ``message``
            attribute gives a reason.
        :return: The access token
        :rtype: ~azure.core.credentials.AccessToken
        """
        if not scopes:
            msg = '"get_token" requires at least one scope'
            raise ValidationException(
                message=msg,
                no_personal_data_message=msg,
                target=ErrorTarget.IDENTITY,
                error_category=ErrorCategory.USER_ERROR,
            )

        try:
            token = await self._acquire_token_silently(*scopes, **kwargs)
            if not token:
                self._last_request_time = int(time.time())
                token = await self._request_token(*scopes, **kwargs)
            elif self._should_refresh(token):
                try:
                    self._last_request_time = int(time.time())
                    token = await self._request_token(*scopes, **kwargs)
                except Exception:  # pylint:disable=broad-except
                    pass
            _LOGGER.log(
                logging.INFO,
                "%s.get_token succeeded",
                self.__class__.__name__,
            )
            return token

        except Exception as ex:
            _LOGGER.log(
                logging.WARNING,
                "%s.get_token failed: %s",
                self.__class__.__name__,
                ex,
                exc_info=_LOGGER.isEnabledFor(logging.DEBUG),
            )
            raise



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Entrypoints for creating AutoML tasks."""
from typing import List, Optional, Union

from azure.ai.ml.entities._builders.base_node import pipeline_node_decorator
from azure.ai.ml.entities._inputs_outputs import Input
from azure.ai.ml.entities._job.automl.tabular import (
    ClassificationJob,
    ForecastingJob,
    ForecastingSettings,
    RegressionJob,
)


@pipeline_node_decorator
def classification(
    *,
    training_data: Input,
    target_column_name: str,
    primary_metric: Optional[str] = None,
    enable_model_explainability: Optional[bool] = None,
    weight_column_name: Optional[str] = None,
    validation_data: Optional[Input] = None,
    validation_data_size: Optional[float] = None,
    n_cross_validations: Optional[Union[str, int]] = None,
    cv_split_column_names: Optional[List[str]] = None,
    test_data: Optional[Input] = None,
    test_data_size: Optional[float] = None,
    **kwargs,
) -> ClassificationJob:
    """Function to create a ClassificationJob.

    A classification job is used to train a model that best predict the class of a data sample.
    Various models are trained using the training data. The model with the best performance on the validation data
    based on the primary metric is selected as the final model.

    :keyword training_data: The training data to be used within the experiment.
            It should contain both training features and a label column (optionally a sample weights column).
    :paramtype training_data: Input
    :keyword target_column_name: The name of the label column.
            This parameter is applicable to ``training_data``, ``validation_data`` and ``test_data`` parameters
    :paramtype target_column_name: str
    :keyword primary_metric: The metric that Automated Machine Learning will optimize for model selection.
            Automated Machine Learning collects more metrics than it can optimize.
            For more information on how metrics are calculated, see
            https://learn.microsoft.com/azure/machine-learning/how-to-configure-auto-train#primary-metric.

            Acceptable values: accuracy, AUC_weighted, norm_macro_recall, average_precision_score_weighted,
            and precision_score_weighted
            Defaults to accuracy
    :type primary_metric: str
    :keyword enable_model_explainability: Whether to enable explaining the best AutoML model at the end of all AutoML
            training iterations.
            The default is None. For more information, see
            `Interpretability: model explanations in automated machine learning
            <https://learn.microsoft.com/azure/machine-learning/how-to-machine-learning-interpretability-automl>`__.
    :paramtype enable_model_explainability: bool
    :keyword weight_column_name: The name of the sample weight column. Automated ML supports a weighted column
            as an input, causing rows in the data to be weighted up or down.
            If the input data is from a pandas.DataFrame which doesn't have column names,
            column indices can be used instead, expressed as integers.

            This parameter is applicable to ``training_data`` and ``validation_data`` parameters
    :paramtype weight_column_name: str
    :keyword validation_data: The validation data to be used within the experiment.
            It should contain both training features and label column (optionally a sample weights column).

            Defaults to None
    :paramtype validation_data: Input
    :keyword validation_data_size: What fraction of the data to hold out for validation when user validation data
            is not specified. This should be between 0.0 and 1.0 non-inclusive.

            Specify ``validation_data`` to provide validation data, otherwise set ``n_cross_validations`` or
            ``validation_data_size`` to extract validation data out of the specified training data.
            For custom cross validation fold, use ``cv_split_column_names``.

            For more information, see
            `Configure data splits and cross-validation in automated machine learning <https://learn.microsoft.com
            /azure/machine-learning/how-to-configure-cross-validation-data-splits>`__.

            Defaults to None
    :paramtype validation_data_size: float
    :keyword n_cross_validations: How many cross validations to perform when user validation data is not specified.

            Specify ``validation_data`` to provide validation data, otherwise set ``n_cross_validations`` or
            ``validation_data_size`` to extract validation data out of the specified training data.
            For custom cross validation fold, use ``cv_split_column_names``.

            For more information, see
            `Configure data splits and cross-validation in automated machine learning <https://learn.microsoft.com
            /azure/machine-learning/how-to-configure-cross-validation-data-splits>`__.

            Defaults to None
    :paramtype n_cross_validations: Union[str, int]
    :keyword cv_split_column_names: List of names of the columns that contain custom cross validation split.
            Each of the CV split columns represents one CV split where each row are either marked
            1 for training or 0 for validation.

            Defaults to None
    :paramtype cv_split_column_names: List[str]
    :keyword test_data: The Model Test feature using test datasets or test data splits is a feature in
            Preview state and might change at any time.
            The test data to be used for a test run that will automatically be started after
            model training is complete. The test run will get predictions using the best model
            and will compute metrics given these predictions.

            If this parameter or the ``test_data_size`` parameter are not specified then
            no test run will be executed automatically after model training is completed.
            Test data should contain both features and label column.
            If ``test_data`` is specified then the ``target_column_name`` parameter must be specified.

            Defaults to None
    :paramtype test_data: Input
    :keyword test_data_size: The Model Test feature using test datasets or test data splits is a feature in
            Preview state and might change at any time.
            What fraction of the training data to hold out for test data for a test run that will
            automatically be started after model training is complete. The test run will get
            predictions using the best model and will compute metrics given these predictions.

            This should be between 0.0 and 1.0 non-inclusive.
            If ``test_data_size`` is specified at the same time as ``validation_data_size``,
            then the test data is split from ``training_data`` before the validation data is split.
            For example, if ``validation_data_size=0.1``, ``test_data_size=0.1`` and the original training data has
            1000 rows, then the test data will have 100 rows, the validation data will contain 90 rows and the
            training data will have 810 rows.

            For regression based tasks, random sampling is used. For classification tasks, stratified sampling
            is used. Forecasting does not currently support specifying a test dataset using a train/test split.

            If this parameter or the ``test_data`` parameter are not specified then
            no test run will be executed automatically after model training is completed.

            Defaults to None
    :paramtype test_data_size: float
    :return: A job object that can be submitted to an Azure ML compute for execution.
    :rtype: ClassificationJob
    """
    classification_job = ClassificationJob(primary_metric=primary_metric, **kwargs)

    classification_job.set_data(
        training_data=training_data,
        target_column_name=target_column_name,
        weight_column_name=weight_column_name,
        validation_data=validation_data,
        validation_data_size=validation_data_size,
        n_cross_validations=n_cross_validations,
        cv_split_column_names=cv_split_column_names,
        test_data=test_data,
        test_data_size=test_data_size,
    )
    classification_job.set_training(enable_model_explainability=enable_model_explainability)

    return classification_job


@pipeline_node_decorator
def regression(
    *,
    training_data: Input,
    target_column_name: str,
    primary_metric: Optional[str] = None,
    enable_model_explainability: Optional[bool] = None,
    weight_column_name: Optional[str] = None,
    validation_data: Optional[Input] = None,
    validation_data_size: Optional[float] = None,
    n_cross_validations: Optional[Union[str, int]] = None,
    cv_split_column_names: Optional[List[str]] = None,
    test_data: Optional[Input] = None,
    test_data_size: Optional[float] = None,
    **kwargs,
) -> RegressionJob:
    """Function to create a Regression Job.

    A regression job is used to train a model to predict continuous values of a target variable from a dataset.
    Various models are trained using the training data. The model with the best performance on the validation data
    based on the primary metric is selected as the final model.


    :keyword training_data: The training data to be used within the experiment.
            It should contain both training features and a label column (optionally a sample weights column).
    :paramtype training_data: Input
    :keyword target_column_name: The name of the label column.
            This parameter is applicable to ``training_data``, ``validation_data`` and ``test_data`` parameters
    :paramtype target_column_name: str
    :keyword primary_metric: The metric that Automated Machine Learning will optimize for model selection.
            Automated Machine Learning collects more metrics than it can optimize.
            For more information on how metrics are calculated, see
            https://learn.microsoft.com/azure/machine-learning/how-to-configure-auto-train#primary-metric.

            Acceptable values: spearman_correlation, r2_score, normalized_mean_absolute_error,
            normalized_root_mean_squared_error.
            Defaults to normalized_root_mean_squared_error
    :type primary_metric: str
    :keyword enable_model_explainability: Whether to enable explaining the best AutoML model at the end of all AutoML
            training iterations.
            The default is None. For more information, see
            `Interpretability: model explanations in automated machine learning
            <https://learn.microsoft.com/azure/machine-learning/how-to-machine-learning-interpretability-automl>`__.
    :paramtype enable_model_explainability: bool
    :keyword weight_column_name: The name of the sample weight column. Automated ML supports a weighted column
            as an input, causing rows in the data to be weighted up or down.
            If the input data is from a pandas.DataFrame which doesn't have column names,
            column indices can be used instead, expressed as integers.

            This parameter is applicable to ``training_data`` and ``validation_data`` parameters
    :paramtype weight_column_name: str
    :keyword validation_data: The validation data to be used within the experiment.
            It should contain both training features and label column (optionally a sample weights column).

            Defaults to None
    :paramtype validation_data: Input
    :keyword validation_data_size: What fraction of the data to hold out for validation when user validation data
            is not specified. This should be between 0.0 and 1.0 non-inclusive.

            Specify ``validation_data`` to provide validation data, otherwise set ``n_cross_validations`` or
            ``validation_data_size`` to extract validation data out of the specified training data.
            For custom cross validation fold, use ``cv_split_column_names``.

            For more information, see
            `Configure data splits and cross-validation in automated machine learning <https://learn.microsoft.com
            /azure/machine-learning/how-to-configure-cross-validation-data-splits>`__.

            Defaults to None
    :paramtype validation_data_size: float
    :keyword n_cross_validations: How many cross validations to perform when user validation data is not specified.

            Specify ``validation_data`` to provide validation data, otherwise set ``n_cross_validations`` or
            ``validation_data_size`` to extract validation data out of the specified training data.
            For custom cross validation fold, use ``cv_split_column_names``.

            For more information, see
            `Configure data splits and cross-validation in automated machine learning <https://learn.microsoft.com
            /azure/machine-learning/how-to-configure-cross-validation-data-splits>`__.

            Defaults to None
    :paramtype n_cross_validations: Union[str, int]
    :keyword cv_split_column_names: List of names of the columns that contain custom cross validation split.
            Each of the CV split columns represents one CV split where each row are either marked
            1 for training or 0 for validation.

            Defaults to None
    :paramtype cv_split_column_names: List[str]
    :keyword test_data: The Model Test feature using test datasets or test data splits is a feature in
            Preview state and might change at any time.
            The test data to be used for a test run that will automatically be started after
            model training is complete. The test run will get predictions using the best model
            and will compute metrics given these predictions.

            If this parameter or the ``test_data_size`` parameter are not specified then
            no test run will be executed automatically after model training is completed.
            Test data should contain both features and label column.
            If ``test_data`` is specified then the ``target_column_name`` parameter must be specified.

            Defaults to None
    :paramtype test_data: Input
    :keyword test_data_size: The Model Test feature using test datasets or test data splits is a feature in
            Preview state and might change at any time.
            What fraction of the training data to hold out for test data for a test run that will
            automatically be started after model training is complete. The test run will get
            predictions using the best model and will compute metrics given these predictions.

            This should be between 0.0 and 1.0 non-inclusive.
            If ``test_data_size`` is specified at the same time as ``validation_data_size``,
            then the test data is split from ``training_data`` before the validation data is split.
            For example, if ``validation_data_size=0.1``, ``test_data_size=0.1`` and the original training data has
            1000 rows, then the test data will have 100 rows, the validation data will contain 90 rows
            and the training data will have 810 rows.

            For regression based tasks, random sampling is used. For classification
            tasks, stratified sampling is used. Forecasting does not currently
            support specifying a test dataset using a train/test split.

            If this parameter or the ``test_data`` parameter are not specified then
            no test run will be executed automatically after model training is completed.

            Defaults to None
    :paramtype test_data_size: float
    :return: A job object that can be submitted to an Azure ML compute for execution.
    :rtype: RegressionJob
    """
    regression_job = RegressionJob(primary_metric=primary_metric, **kwargs)
    regression_job.set_data(
        training_data=training_data,
        target_column_name=target_column_name,
        weight_column_name=weight_column_name,
        validation_data=validation_data,
        validation_data_size=validation_data_size,
        n_cross_validations=n_cross_validations,
        cv_split_column_names=cv_split_column_names,
        test_data=test_data,
        test_data_size=test_data_size,
    )
    regression_job.set_training(enable_model_explainability=enable_model_explainability)

    return regression_job


@pipeline_node_decorator
def forecasting(
    *,
    training_data: Input,
    target_column_name: str,
    primary_metric: Optional[str] = None,
    enable_model_explainability: Optional[bool] = None,
    weight_column_name: Optional[str] = None,
    validation_data: Optional[Input] = None,
    validation_data_size: Optional[float] = None,
    n_cross_validations: Optional[Union[str, int]] = None,
    cv_split_column_names: Optional[List[str]] = None,
    test_data: Optional[Input] = None,
    test_data_size: Optional[float] = None,
    forecasting_settings: Optional[ForecastingSettings] = None,
    **kwargs,
) -> ForecastingJob:
    """Function to create a Forecasting job.

    A forecasting task is used to predict target values for a future time period based on the historical data.
    Various models are trained using the training data. The model with the best performance on the validation data
    based on the primary metric is selected as the final model.

    :keyword training_data: The training data to be used within the experiment.
            It should contain both training features and a label column (optionally a sample weights column).
    :paramtype training_data: Input
    :keyword target_column_name: The name of the label column.
            This parameter is applicable to ``training_data``, ``validation_data`` and ``test_data`` parameters
    :paramtype target_column_name: str
    :keyword primary_metric: The metric that Automated Machine Learning will optimize for model selection.
            Automated Machine Learning collects more metrics than it can optimize.
            For more information on how metrics are calculated, see
            https://learn.microsoft.com/azure/machine-learning/how-to-configure-auto-train#primary-metric.

            Acceptable values: r2_score, normalized_mean_absolute_error, normalized_root_mean_squared_error
            Defaults to normalized_root_mean_squared_error
    :type primary_metric: str
    :keyword enable_model_explainability: Whether to enable explaining the best AutoML model at the end of all AutoML
            training iterations.
            The default is None. For more information, see
            `Interpretability: model explanations in automated machine learning
            <https://learn.microsoft.com/azure/machine-learning/how-to-machine-learning-interpretability-automl>`__.
    :paramtype enable_model_explainability: bool
    :keyword weight_column_name: The name of the sample weight column. Automated ML supports a weighted column
            as an input, causing rows in the data to be weighted up or down.
            If the input data is from a pandas.DataFrame which doesn't have column names,
            column indices can be used instead, expressed as integers.

            This parameter is applicable to ``training_data`` and ``validation_data`` parameters
    :paramtype weight_column_name: str
    :keyword validation_data: The validation data to be used within the experiment.
            It should contain both training features and label column (optionally a sample weights column).

            Defaults to None
    :paramtype validation_data: Input
    :keyword validation_data_size: What fraction of the data to hold out for validation when user validation data
            is not specified. This should be between 0.0 and 1.0 non-inclusive.

            Specify ``validation_data`` to provide validation data, otherwise set ``n_cross_validations`` or
            ``validation_data_size`` to extract validation data out of the specified training data.
            For custom cross validation fold, use ``cv_split_column_names``.

            For more information, see
            `Configure data splits and cross-validation in automated machine learning <https://learn.microsoft.com
            /azure/machine-learning/how-to-configure-cross-validation-data-splits>`__.

            Defaults to None
    :paramtype validation_data_size: float
    :keyword n_cross_validations: How many cross validations to perform when user validation data is not specified.

            Specify ``validation_data`` to provide validation data, otherwise set ``n_cross_validations`` or
            ``validation_data_size`` to extract validation data out of the specified training data.
            For custom cross validation fold, use ``cv_split_column_names``.

            For more information, see
            `Configure data splits and cross-validation in automated machine learning <https://learn.microsoft.com
            /azure/machine-learning/how-to-configure-cross-validation-data-splits>`__.

            Defaults to None
    :paramtype n_cross_validations: Union[str, int]
    :keyword cv_split_column_names: List of names of the columns that contain custom cross validation split.
            Each of the CV split columns represents one CV split where each row are either marked
            1 for training or 0 for validation.

            Defaults to None
    :paramtype cv_split_column_names: List[str]
    :keyword test_data: The Model Test feature using test datasets or test data splits is a feature in
            Preview state and might change at any time.
            The test data to be used for a test run that will automatically be started after
            model training is complete. The test run will get predictions using the best model
            and will compute metrics given these predictions.

            If this parameter or the ``test_data_size`` parameter are not specified then
            no test run will be executed automatically after model training is completed.
            Test data should contain both features and label column.
            If ``test_data`` is specified then the ``target_column_name`` parameter must be specified.

            Defaults to None
    :paramtype test_data: Input
    :keyword test_data_size: The Model Test feature using test datasets or test data splits is a feature in
            Preview state and might change at any time.
            What fraction of the training data to hold out for test data for a test run that will
            automatically be started after model training is complete. The test run will get
            predictions using the best model and will compute metrics given these predictions.

            This should be between 0.0 and 1.0 non-inclusive.
            If ``test_data_size`` is specified at the same time as ``validation_data_size``,
            then the test data is split from ``training_data`` before the validation data is split.
            For example, if ``validation_data_size=0.1``, ``test_data_size=0.1`` and the original training data
            has 1000 rows, then the test data will have 100 rows, the validation data will contain 90 rows
            and the training data will have 810 rows.

            For regression based tasks, random sampling is used. For classification
            tasks, stratified sampling is used. Forecasting does not currently
            support specifying a test dataset using a train/test split.

            If this parameter or the ``test_data`` parameter are not specified then
            no test run will be executed automatically after model training is completed.

            Defaults to None
    :paramtype test_data_size: float
    :keyword forecasting_settings: The settings for the forecasting task
    :paramtype forecasting_settings: ForecastingSettings
    :return: A job object that can be submitted to an Azure ML compute for execution.
    :rtype: ForecastingJob
    """
    forecast_job = ForecastingJob(
        primary_metric=primary_metric,
        forecasting_settings=forecasting_settings,
        **kwargs,
    )
    forecast_job.set_data(
        training_data=training_data,
        target_column_name=target_column_name,
        weight_column_name=weight_column_name,
        validation_data=validation_data,
        validation_data_size=validation_data_size,
        n_cross_validations=n_cross_validations,
        cv_split_column_names=cv_split_column_names,
        test_data=test_data,
        test_data_size=test_data_size,
    )
    forecast_job.set_training(enable_model_explainability=enable_model_explainability)

    return forecast_job



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

import logging
from typing import Any

from marshmallow import fields, post_load

from azure.ai.ml._schema.core.fields import NestedField, UnionField
from azure.ai.ml._schema.job.input_output_entry import OutputSchema
from azure.ai.ml._schema.job.spark_job import SparkJobSchema

module_logger = logging.getLogger(__name__)


class PipelineSparkJobSchema(SparkJobSchema):
    outputs = fields.Dict(
        keys=fields.Str(),
        values=UnionField([NestedField(OutputSchema), fields.Str()], allow_none=True),
    )

    @post_load
    def make(self, data: Any, **kwargs: Any):
        from azure.ai.ml.entities._job.spark_job import SparkJob

        return SparkJob(**data)



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from ._azure_machine_learning_services_enums import (
    ActionType,
    AllocationState,
    ApplicationSharingPolicy,
    AssetProvisioningState,
    AutoRebuildSetting,
    Autosave,
    BatchLoggingLevel,
    BatchOutputAction,
    BillingCurrency,
    BlockedTransformers,
    Caching,
    CategoricalDataDriftMetric,
    CategoricalDataQualityMetric,
    CategoricalPredictionDriftMetric,
    ClassificationModels,
    ClassificationMultilabelPrimaryMetrics,
    ClassificationPrimaryMetrics,
    ClusterPurpose,
    ComputeInstanceAuthorizationType,
    ComputeInstanceState,
    ComputePowerAction,
    ComputeRecurrenceFrequency,
    ComputeTriggerType,
    ComputeType,
    ComputeWeekDay,
    ConnectionAuthType,
    ConnectionCategory,
    ContainerType,
    CreatedByType,
    CredentialsType,
    DataAvailabilityStatus,
    DataReferenceCredentialType,
    DatastoreType,
    DataType,
    DeploymentProvisioningState,
    DiagnoseResultLevel,
    DistributionType,
    EarlyTerminationPolicyType,
    EgressPublicNetworkAccessType,
    EmailNotificationEnableType,
    EncryptionStatus,
    EndpointAuthMode,
    EndpointComputeType,
    EndpointProvisioningState,
    EndpointServiceConnectionStatus,
    EnvironmentType,
    EnvironmentVariableType,
    FeatureAttributionMetric,
    FeatureDataType,
    FeatureImportanceMode,
    FeatureLags,
    FeaturizationMode,
    ForecastHorizonMode,
    ForecastingModels,
    ForecastingPrimaryMetrics,
    Goal,
    IdentityConfigurationType,
    ImageType,
    InputDeliveryMode,
    InstanceSegmentationPrimaryMetrics,
    IsolationMode,
    JobInputType,
    JobLimitsType,
    JobOutputType,
    JobStatus,
    JobTier,
    JobType,
    KeyType,
    LearningRateScheduler,
    ListViewType,
    LoadBalancerType,
    LogVerbosity,
    ManagedNetworkStatus,
    ManagedServiceIdentityType,
    MaterializationStoreType,
    ModelSize,
    ModelTaskType,
    MonitorComputeIdentityType,
    MonitorComputeType,
    MonitoringFeatureDataType,
    MonitoringFeatureFilterType,
    MonitoringInputDataType,
    MonitoringNotificationType,
    MonitoringSignalType,
    MountAction,
    MountState,
    NCrossValidationsMode,
    Network,
    NodeState,
    NodesValueType,
    NumericalDataDriftMetric,
    NumericalDataQualityMetric,
    NumericalPredictionDriftMetric,
    ObjectDetectionPrimaryMetrics,
    OperatingSystemType,
    OperationName,
    OperationStatus,
    OperationTrigger,
    OrderString,
    Origin,
    OsType,
    OutputDeliveryMode,
    PendingUploadCredentialType,
    PendingUploadType,
    PrivateEndpointConnectionProvisioningState,
    PrivateEndpointServiceConnectionStatus,
    Protocol,
    ProvisioningState,
    ProvisioningStatus,
    PublicNetworkAccess,
    PublicNetworkAccessType,
    QuotaUnit,
    RandomSamplingAlgorithmRule,
    RecurrenceFrequency,
    ReferenceType,
    RegressionModels,
    RegressionPrimaryMetrics,
    RemoteLoginPortPublicAccess,
    RuleAction,
    RuleCategory,
    RuleStatus,
    RuleType,
    SamplingAlgorithmType,
    ScaleType,
    ScheduleActionType,
    ScheduleListViewType,
    ScheduleProvisioningState,
    ScheduleProvisioningStatus,
    ScheduleStatus,
    SeasonalityMode,
    SecretsType,
    ServiceDataAccessAuthIdentity,
    ShortSeriesHandlingConfiguration,
    SkuScaleType,
    SkuTier,
    SourceType,
    SshPublicAccess,
    SslConfigStatus,
    StackMetaLearnerType,
    Status,
    StochasticOptimizer,
    StorageAccountType,
    TargetAggregationFunction,
    TargetLagsMode,
    TargetRollingWindowSizeMode,
    TaskType,
    TriggerType,
    UnderlyingResourceAction,
    UnitOfMeasure,
    UsageUnit,
    UseStl,
    ValidationMetricType,
    ValueFormat,
    VMPriceOSType,
    VmPriority,
    VMTier,
    VolumeDefinitionType,
    WebhookType,
    WeekDay,
)
from ._models_py3 import (
    AKS,
    AccountKeyDatastoreCredentials,
    AccountKeyDatastoreSecrets,
    AcrDetails,
    AksComputeSecrets,
    AksComputeSecretsProperties,
    AksNetworkingConfiguration,
    AKSSchema,
    AKSSchemaProperties,
    AllFeatures,
    AllNodes,
    AmlCompute,
    AmlComputeNodeInformation,
    AmlComputeNodesInformation,
    AmlComputeProperties,
    AmlComputeSchema,
    AmlToken,
    AmlTokenComputeIdentity,
    AmlUserFeature,
    AnonymousAccessCredential,
    ArmResourceId,
    AssetBase,
    AssetContainer,
    AssetJobInput,
    AssetJobOutput,
    AssetReferenceBase,
    AssignedUser,
    AutoForecastHorizon,
    AutoMLJob,
    AutoMLVertical,
    AutoNCrossValidations,
    AutoPauseProperties,
    AutoScaleProperties,
    AutoSeasonality,
    AutoTargetLags,
    AutoTargetRollingWindowSize,
    AzureBlobDatastore,
    AzureDataLakeGen1Datastore,
    AzureDataLakeGen2Datastore,
    AzureDatastore,
    AzureDevOpsWebhook,
    AzureFileDatastore,
    BanditPolicy,
    BatchDeployment,
    BatchDeploymentProperties,
    BatchDeploymentTrackedResourceArmPaginatedResult,
    BatchEndpoint,
    BatchEndpointDefaults,
    BatchEndpointProperties,
    BatchEndpointTrackedResourceArmPaginatedResult,
    BatchRetrySettings,
    BayesianSamplingAlgorithm,
    BindOptions,
    BlobReferenceForConsumptionDto,
    BuildContext,
    CategoricalDataDriftMetricThreshold,
    CategoricalDataQualityMetricThreshold,
    CategoricalPredictionDriftMetricThreshold,
    CertificateDatastoreCredentials,
    CertificateDatastoreSecrets,
    Classification,
    ClassificationTrainingSettings,
    ClusterUpdateParameters,
    CodeConfiguration,
    CodeContainer,
    CodeContainerProperties,
    CodeContainerResourceArmPaginatedResult,
    CodeVersion,
    CodeVersionProperties,
    CodeVersionResourceArmPaginatedResult,
    ColumnTransformer,
    CommandJob,
    CommandJobLimits,
    ComponentContainer,
    ComponentContainerProperties,
    ComponentContainerResourceArmPaginatedResult,
    ComponentVersion,
    ComponentVersionProperties,
    ComponentVersionResourceArmPaginatedResult,
    Compute,
    ComputeInstance,
    ComputeInstanceApplication,
    ComputeInstanceConnectivityEndpoints,
    ComputeInstanceContainer,
    ComputeInstanceCreatedBy,
    ComputeInstanceDataDisk,
    ComputeInstanceDataMount,
    ComputeInstanceEnvironmentInfo,
    ComputeInstanceLastOperation,
    ComputeInstanceProperties,
    ComputeInstanceSchema,
    ComputeInstanceSshSettings,
    ComputeInstanceVersion,
    ComputeRecurrenceSchedule,
    ComputeResource,
    ComputeResourceSchema,
    ComputeSchedules,
    ComputeSecrets,
    ComputeStartStopSchedule,
    ContainerResourceRequirements,
    ContainerResourceSettings,
    CosmosDbSettings,
    CreateMonitorAction,
    Cron,
    CronTrigger,
    CustomForecastHorizon,
    CustomMetricThreshold,
    CustomModelJobInput,
    CustomModelJobOutput,
    CustomMonitoringSignal,
    CustomNCrossValidations,
    CustomSeasonality,
    CustomService,
    CustomTargetLags,
    CustomTargetRollingWindowSize,
    Databricks,
    DatabricksComputeSecrets,
    DatabricksComputeSecretsProperties,
    DatabricksProperties,
    DatabricksSchema,
    DataContainer,
    DataContainerProperties,
    DataContainerResourceArmPaginatedResult,
    DataDriftMetricThresholdBase,
    DataDriftMonitoringSignal,
    DataFactory,
    DataLakeAnalytics,
    DataLakeAnalyticsSchema,
    DataLakeAnalyticsSchemaProperties,
    DataPathAssetReference,
    DataQualityMetricThresholdBase,
    DataQualityMonitoringSignal,
    DataReferenceCredential,
    Datastore,
    DatastoreCredentials,
    DatastoreProperties,
    DatastoreResourceArmPaginatedResult,
    DatastoreSecrets,
    DataVersionBase,
    DataVersionBaseProperties,
    DataVersionBaseResourceArmPaginatedResult,
    DefaultScaleSettings,
    DeploymentLogs,
    DeploymentLogsRequest,
    DeploymentResourceConfiguration,
    DestinationAsset,
    DiagnoseRequestProperties,
    DiagnoseResponseResult,
    DiagnoseResponseResultValue,
    DiagnoseResult,
    DiagnoseWorkspaceParameters,
    DistributionConfiguration,
    Docker,
    DockerCredential,
    EarlyTerminationPolicy,
    EncryptionKeyVaultProperties,
    EncryptionProperty,
    Endpoint,
    EndpointAuthKeys,
    EndpointAuthToken,
    EndpointDeploymentPropertiesBase,
    EndpointPropertiesBase,
    EndpointScheduleAction,
    EnvironmentContainer,
    EnvironmentContainerProperties,
    EnvironmentContainerResourceArmPaginatedResult,
    EnvironmentVariable,
    EnvironmentVersion,
    EnvironmentVersionProperties,
    EnvironmentVersionResourceArmPaginatedResult,
    ErrorAdditionalInfo,
    ErrorDetail,
    ErrorResponse,
    EstimatedVMPrice,
    EstimatedVMPrices,
    ExternalFQDNResponse,
    Feature,
    FeatureAttributionDriftMonitoringSignal,
    FeatureAttributionMetricThreshold,
    FeatureImportanceSettings,
    FeatureProperties,
    FeatureResourceArmPaginatedResult,
    FeaturesetContainer,
    FeaturesetContainerProperties,
    FeaturesetContainerResourceArmPaginatedResult,
    FeaturesetSpecification,
    FeaturesetVersion,
    FeaturesetVersionBackfillRequest,
    FeaturesetVersionBackfillResponse,
    FeaturesetVersionProperties,
    FeaturesetVersionResourceArmPaginatedResult,
    FeaturestoreEntityContainer,
    FeaturestoreEntityContainerProperties,
    FeaturestoreEntityContainerResourceArmPaginatedResult,
    FeaturestoreEntityVersion,
    FeaturestoreEntityVersionProperties,
    FeaturestoreEntityVersionResourceArmPaginatedResult,
    FeatureSubset,
    FeatureWindow,
    FeaturizationSettings,
    FixedInputData,
    FlavorData,
    ForecastHorizon,
    Forecasting,
    ForecastingSettings,
    ForecastingTrainingSettings,
    FQDNEndpoint,
    FQDNEndpointDetail,
    FQDNEndpoints,
    FQDNEndpointsProperties,
    FqdnOutboundRule,
    GetBlobReferenceForConsumptionDto,
    GetBlobReferenceSASRequestDto,
    GetBlobReferenceSASResponseDto,
    GridSamplingAlgorithm,
    HDInsight,
    HDInsightProperties,
    HDInsightSchema,
    IdAssetReference,
    IdentityConfiguration,
    IdentityForCmk,
    IdleShutdownSetting,
    Image,
    ImageClassification,
    ImageClassificationBase,
    ImageClassificationMultilabel,
    ImageInstanceSegmentation,
    ImageLimitSettings,
    ImageMetadata,
    ImageModelDistributionSettings,
    ImageModelDistributionSettingsClassification,
    ImageModelDistributionSettingsObjectDetection,
    ImageModelSettings,
    ImageModelSettingsClassification,
    ImageModelSettingsObjectDetection,
    ImageObjectDetection,
    ImageObjectDetectionBase,
    ImageSweepSettings,
    ImageVertical,
    IndexColumn,
    InferenceContainerProperties,
    InstanceTypeSchema,
    InstanceTypeSchemaResources,
    JobBase,
    JobBaseProperties,
    JobBaseResourceArmPaginatedResult,
    JobInput,
    JobLimits,
    JobOutput,
    JobResourceConfiguration,
    JobScheduleAction,
    JobService,
    Kubernetes,
    KubernetesOnlineDeployment,
    KubernetesProperties,
    KubernetesSchema,
    ListAmlUserFeatureResult,
    ListNotebookKeysResult,
    ListStorageAccountKeysResult,
    ListUsagesResult,
    ListWorkspaceKeysResult,
    ListWorkspaceQuotas,
    LiteralJobInput,
    ManagedComputeIdentity,
    ManagedIdentity,
    ManagedIdentityAuthTypeWorkspaceConnectionProperties,
    ManagedIdentityCredential,
    ManagedNetworkProvisionOptions,
    ManagedNetworkProvisionStatus,
    ManagedNetworkSettings,
    ManagedOnlineDeployment,
    ManagedServiceIdentity,
    MaterializationComputeResource,
    MaterializationSettings,
    MedianStoppingPolicy,
    MLFlowModelJobInput,
    MLFlowModelJobOutput,
    MLTableData,
    MLTableJobInput,
    MLTableJobOutput,
    ModelContainer,
    ModelContainerProperties,
    ModelContainerResourceArmPaginatedResult,
    ModelVersion,
    ModelVersionProperties,
    ModelVersionResourceArmPaginatedResult,
    MonitorComputeConfigurationBase,
    MonitorComputeIdentityBase,
    MonitorDefinition,
    MonitorEmailNotificationSettings,
    MonitoringFeatureFilterBase,
    MonitoringInputDataBase,
    MonitoringSignalBase,
    MonitoringTarget,
    MonitoringThreshold,
    MonitorNotificationSettings,
    MonitorServerlessSparkCompute,
    Mpi,
    NCrossValidations,
    NlpVertical,
    NlpVerticalFeaturizationSettings,
    NlpVerticalLimitSettings,
    Nodes,
    NodeStateCounts,
    NoneAuthTypeWorkspaceConnectionProperties,
    NoneDatastoreCredentials,
    NotebookAccessTokenResult,
    NotebookPreparationError,
    NotebookResourceInfo,
    NotificationSetting,
    NumericalDataDriftMetricThreshold,
    NumericalDataQualityMetricThreshold,
    NumericalPredictionDriftMetricThreshold,
    Objective,
    OnlineDeployment,
    OnlineDeploymentProperties,
    OnlineDeploymentTrackedResourceArmPaginatedResult,
    OnlineEndpoint,
    OnlineEndpointProperties,
    OnlineEndpointTrackedResourceArmPaginatedResult,
    OnlineRequestSettings,
    OnlineScaleSettings,
    Operation,
    OperationDisplay,
    OperationListResult,
    OutboundRule,
    OutboundRuleBasicResource,
    OutboundRuleListResult,
    OutputPathAssetReference,
    PaginatedComputeResourcesList,
    PartialBatchDeployment,
    PartialBatchDeploymentPartialMinimalTrackedResourceWithProperties,
    PartialManagedServiceIdentity,
    PartialMinimalTrackedResource,
    PartialMinimalTrackedResourceWithIdentity,
    PartialMinimalTrackedResourceWithSku,
    PartialRegistryPartialTrackedResource,
    PartialSku,
    Password,
    PATAuthTypeWorkspaceConnectionProperties,
    PendingUploadCredentialDto,
    PendingUploadRequestDto,
    PendingUploadResponseDto,
    PersonalComputeInstanceSettings,
    PipelineJob,
    PredictionDriftMetricThresholdBase,
    PredictionDriftMonitoringSignal,
    PrivateEndpoint,
    PrivateEndpointConnection,
    PrivateEndpointConnectionListResult,
    PrivateEndpointDestination,
    PrivateEndpointOutboundRule,
    PrivateEndpointResource,
    PrivateLinkResource,
    PrivateLinkResourceListResult,
    PrivateLinkServiceConnectionState,
    ProbeSettings,
    ProxyResource,
    PyTorch,
    QueueSettings,
    QuotaBaseProperties,
    QuotaUpdateParameters,
    RandomSamplingAlgorithm,
    Recurrence,
    RecurrenceSchedule,
    RecurrenceTrigger,
    RegenerateEndpointKeysRequest,
    Registry,
    RegistryListCredentialsResult,
    RegistryPartialManagedServiceIdentity,
    RegistryPrivateEndpointConnection,
    RegistryPrivateLinkServiceConnectionState,
    RegistryRegionArmDetails,
    RegistryTrackedResourceArmPaginatedResult,
    Regression,
    RegressionTrainingSettings,
    Resource,
    ResourceBase,
    ResourceConfiguration,
    ResourceId,
    ResourceName,
    ResourceQuota,
    RollingInputData,
    Route,
    SamplingAlgorithm,
    SASAuthTypeWorkspaceConnectionProperties,
    SASCredential,
    SASCredentialDto,
    SasDatastoreCredentials,
    SasDatastoreSecrets,
    ScaleSettings,
    ScaleSettingsInformation,
    Schedule,
    ScheduleActionBase,
    ScheduleBase,
    ScheduleProperties,
    ScheduleResourceArmPaginatedResult,
    ScriptReference,
    ScriptsToExecute,
    Seasonality,
    ServerlessComputeSettings,
    ServiceManagedResourcesSettings,
    ServicePrincipalDatastoreCredentials,
    ServicePrincipalDatastoreSecrets,
    ServiceTagDestination,
    ServiceTagOutboundRule,
    SetupScripts,
    SharedPrivateLinkResource,
    Sku,
    SkuCapacity,
    SkuResource,
    SkuResourceArmPaginatedResult,
    SkuSetting,
    SslConfiguration,
    StackEnsembleSettings,
    StaticInputData,
    StorageAccountDetails,
    SweepJob,
    SweepJobLimits,
    SynapseSpark,
    SynapseSparkProperties,
    SystemCreatedAcrAccount,
    SystemCreatedStorageAccount,
    SystemData,
    SystemService,
    TableVertical,
    TableVerticalFeaturizationSettings,
    TableVerticalLimitSettings,
    TargetLags,
    TargetRollingWindowSize,
    TargetUtilizationScaleSettings,
    TensorFlow,
    TextClassification,
    TextClassificationMultilabel,
    TextNer,
    TmpfsOptions,
    TopNFeaturesByAttribution,
    TrackedResource,
    TrainingSettings,
    TrialComponent,
    TriggerBase,
    TritonModelJobInput,
    TritonModelJobOutput,
    TruncationSelectionPolicy,
    UpdateWorkspaceQuotas,
    UpdateWorkspaceQuotasResult,
    UriFileDataVersion,
    UriFileJobInput,
    UriFileJobOutput,
    UriFolderDataVersion,
    UriFolderJobInput,
    UriFolderJobOutput,
    Usage,
    UsageName,
    UserAccountCredentials,
    UserAssignedIdentity,
    UserCreatedAcrAccount,
    UserCreatedStorageAccount,
    UserIdentity,
    UsernamePasswordAuthTypeWorkspaceConnectionProperties,
    VirtualMachine,
    VirtualMachineImage,
    VirtualMachineSchema,
    VirtualMachineSchemaProperties,
    VirtualMachineSecrets,
    VirtualMachineSecretsSchema,
    VirtualMachineSize,
    VirtualMachineSizeListResult,
    VirtualMachineSshCredentials,
    VolumeDefinition,
    VolumeOptions,
    Webhook,
    Workspace,
    WorkspaceConnectionManagedIdentity,
    WorkspaceConnectionPersonalAccessToken,
    WorkspaceConnectionPropertiesV2,
    WorkspaceConnectionPropertiesV2BasicResource,
    WorkspaceConnectionPropertiesV2BasicResourceArmPaginatedResult,
    WorkspaceConnectionSharedAccessSignature,
    WorkspaceConnectionUsernamePassword,
    WorkspaceListResult,
    WorkspaceUpdateParameters,
)
from ._patch import *  # pylint: disable=unused-wildcard-import
from ._patch import __all__ as _patch_all
from ._patch import patch_sdk as _patch_sdk

__all__ = [
    "AKS",
    "AKSSchema",
    "AKSSchemaProperties",
    "AccountKeyDatastoreCredentials",
    "AccountKeyDatastoreSecrets",
    "AcrDetails",
    "AksComputeSecrets",
    "AksComputeSecretsProperties",
    "AksNetworkingConfiguration",
    "AllFeatures",
    "AllNodes",
    "AmlCompute",
    "AmlComputeNodeInformation",
    "AmlComputeNodesInformation",
    "AmlComputeProperties",
    "AmlComputeSchema",
    "AmlToken",
    "AmlTokenComputeIdentity",
    "AmlUserFeature",
    "AnonymousAccessCredential",
    "ArmResourceId",
    "AssetBase",
    "AssetContainer",
    "AssetJobInput",
    "AssetJobOutput",
    "AssetReferenceBase",
    "AssignedUser",
    "AutoForecastHorizon",
    "AutoMLJob",
    "AutoMLVertical",
    "AutoNCrossValidations",
    "AutoPauseProperties",
    "AutoScaleProperties",
    "AutoSeasonality",
    "AutoTargetLags",
    "AutoTargetRollingWindowSize",
    "AzureBlobDatastore",
    "AzureDataLakeGen1Datastore",
    "AzureDataLakeGen2Datastore",
    "AzureDatastore",
    "AzureDevOpsWebhook",
    "AzureFileDatastore",
    "BanditPolicy",
    "BatchDeployment",
    "BatchDeploymentProperties",
    "BatchDeploymentTrackedResourceArmPaginatedResult",
    "BatchEndpoint",
    "BatchEndpointDefaults",
    "BatchEndpointProperties",
    "BatchEndpointTrackedResourceArmPaginatedResult",
    "BatchRetrySettings",
    "BayesianSamplingAlgorithm",
    "BindOptions",
    "BlobReferenceForConsumptionDto",
    "BuildContext",
    "CategoricalDataDriftMetricThreshold",
    "CategoricalDataQualityMetricThreshold",
    "CategoricalPredictionDriftMetricThreshold",
    "CertificateDatastoreCredentials",
    "CertificateDatastoreSecrets",
    "Classification",
    "ClassificationTrainingSettings",
    "ClusterUpdateParameters",
    "CodeConfiguration",
    "CodeContainer",
    "CodeContainerProperties",
    "CodeContainerResourceArmPaginatedResult",
    "CodeVersion",
    "CodeVersionProperties",
    "CodeVersionResourceArmPaginatedResult",
    "ColumnTransformer",
    "CommandJob",
    "CommandJobLimits",
    "ComponentContainer",
    "ComponentContainerProperties",
    "ComponentContainerResourceArmPaginatedResult",
    "ComponentVersion",
    "ComponentVersionProperties",
    "ComponentVersionResourceArmPaginatedResult",
    "Compute",
    "ComputeInstance",
    "ComputeInstanceApplication",
    "ComputeInstanceConnectivityEndpoints",
    "ComputeInstanceContainer",
    "ComputeInstanceCreatedBy",
    "ComputeInstanceDataDisk",
    "ComputeInstanceDataMount",
    "ComputeInstanceEnvironmentInfo",
    "ComputeInstanceLastOperation",
    "ComputeInstanceProperties",
    "ComputeInstanceSchema",
    "ComputeInstanceSshSettings",
    "ComputeInstanceVersion",
    "ComputeRecurrenceSchedule",
    "ComputeResource",
    "ComputeResourceSchema",
    "ComputeSchedules",
    "ComputeSecrets",
    "ComputeStartStopSchedule",
    "ContainerResourceRequirements",
    "ContainerResourceSettings",
    "CosmosDbSettings",
    "CreateMonitorAction",
    "Cron",
    "CronTrigger",
    "CustomForecastHorizon",
    "CustomMetricThreshold",
    "CustomModelJobInput",
    "CustomModelJobOutput",
    "CustomMonitoringSignal",
    "CustomNCrossValidations",
    "CustomSeasonality",
    "CustomService",
    "CustomTargetLags",
    "CustomTargetRollingWindowSize",
    "DataContainer",
    "DataContainerProperties",
    "DataContainerResourceArmPaginatedResult",
    "DataDriftMetricThresholdBase",
    "DataDriftMonitoringSignal",
    "DataFactory",
    "DataLakeAnalytics",
    "DataLakeAnalyticsSchema",
    "DataLakeAnalyticsSchemaProperties",
    "DataPathAssetReference",
    "DataQualityMetricThresholdBase",
    "DataQualityMonitoringSignal",
    "DataReferenceCredential",
    "DataVersionBase",
    "DataVersionBaseProperties",
    "DataVersionBaseResourceArmPaginatedResult",
    "Databricks",
    "DatabricksComputeSecrets",
    "DatabricksComputeSecretsProperties",
    "DatabricksProperties",
    "DatabricksSchema",
    "Datastore",
    "DatastoreCredentials",
    "DatastoreProperties",
    "DatastoreResourceArmPaginatedResult",
    "DatastoreSecrets",
    "DefaultScaleSettings",
    "DeploymentLogs",
    "DeploymentLogsRequest",
    "DeploymentResourceConfiguration",
    "DestinationAsset",
    "DiagnoseRequestProperties",
    "DiagnoseResponseResult",
    "DiagnoseResponseResultValue",
    "DiagnoseResult",
    "DiagnoseWorkspaceParameters",
    "DistributionConfiguration",
    "Docker",
    "DockerCredential",
    "EarlyTerminationPolicy",
    "EncryptionKeyVaultProperties",
    "EncryptionProperty",
    "Endpoint",
    "EndpointAuthKeys",
    "EndpointAuthToken",
    "EndpointDeploymentPropertiesBase",
    "EndpointPropertiesBase",
    "EndpointScheduleAction",
    "EnvironmentContainer",
    "EnvironmentContainerProperties",
    "EnvironmentContainerResourceArmPaginatedResult",
    "EnvironmentVariable",
    "EnvironmentVersion",
    "EnvironmentVersionProperties",
    "EnvironmentVersionResourceArmPaginatedResult",
    "ErrorAdditionalInfo",
    "ErrorDetail",
    "ErrorResponse",
    "EstimatedVMPrice",
    "EstimatedVMPrices",
    "ExternalFQDNResponse",
    "FQDNEndpoint",
    "FQDNEndpointDetail",
    "FQDNEndpoints",
    "FQDNEndpointsProperties",
    "Feature",
    "FeatureAttributionDriftMonitoringSignal",
    "FeatureAttributionMetricThreshold",
    "FeatureImportanceSettings",
    "FeatureProperties",
    "FeatureResourceArmPaginatedResult",
    "FeatureSubset",
    "FeatureWindow",
    "FeaturesetContainer",
    "FeaturesetContainerProperties",
    "FeaturesetContainerResourceArmPaginatedResult",
    "FeaturesetSpecification",
    "FeaturesetVersion",
    "FeaturesetVersionBackfillRequest",
    "FeaturesetVersionBackfillResponse",
    "FeaturesetVersionProperties",
    "FeaturesetVersionResourceArmPaginatedResult",
    "FeaturestoreEntityContainer",
    "FeaturestoreEntityContainerProperties",
    "FeaturestoreEntityContainerResourceArmPaginatedResult",
    "FeaturestoreEntityVersion",
    "FeaturestoreEntityVersionProperties",
    "FeaturestoreEntityVersionResourceArmPaginatedResult",
    "FeaturizationSettings",
    "FixedInputData",
    "FlavorData",
    "ForecastHorizon",
    "Forecasting",
    "ForecastingSettings",
    "ForecastingTrainingSettings",
    "FqdnOutboundRule",
    "GetBlobReferenceForConsumptionDto",
    "GetBlobReferenceSASRequestDto",
    "GetBlobReferenceSASResponseDto",
    "GridSamplingAlgorithm",
    "HDInsight",
    "HDInsightProperties",
    "HDInsightSchema",
    "IdAssetReference",
    "IdentityConfiguration",
    "IdentityForCmk",
    "IdleShutdownSetting",
    "Image",
    "ImageClassification",
    "ImageClassificationBase",
    "ImageClassificationMultilabel",
    "ImageInstanceSegmentation",
    "ImageLimitSettings",
    "ImageMetadata",
    "ImageModelDistributionSettings",
    "ImageModelDistributionSettingsClassification",
    "ImageModelDistributionSettingsObjectDetection",
    "ImageModelSettings",
    "ImageModelSettingsClassification",
    "ImageModelSettingsObjectDetection",
    "ImageObjectDetection",
    "ImageObjectDetectionBase",
    "ImageSweepSettings",
    "ImageVertical",
    "IndexColumn",
    "InferenceContainerProperties",
    "InstanceTypeSchema",
    "InstanceTypeSchemaResources",
    "JobBase",
    "JobBaseProperties",
    "JobBaseResourceArmPaginatedResult",
    "JobInput",
    "JobLimits",
    "JobOutput",
    "JobResourceConfiguration",
    "JobScheduleAction",
    "JobService",
    "Kubernetes",
    "KubernetesOnlineDeployment",
    "KubernetesProperties",
    "KubernetesSchema",
    "ListAmlUserFeatureResult",
    "ListNotebookKeysResult",
    "ListStorageAccountKeysResult",
    "ListUsagesResult",
    "ListWorkspaceKeysResult",
    "ListWorkspaceQuotas",
    "LiteralJobInput",
    "MLFlowModelJobInput",
    "MLFlowModelJobOutput",
    "MLTableData",
    "MLTableJobInput",
    "MLTableJobOutput",
    "ManagedComputeIdentity",
    "ManagedIdentity",
    "ManagedIdentityAuthTypeWorkspaceConnectionProperties",
    "ManagedIdentityCredential",
    "ManagedNetworkProvisionOptions",
    "ManagedNetworkProvisionStatus",
    "ManagedNetworkSettings",
    "ManagedOnlineDeployment",
    "ManagedServiceIdentity",
    "MaterializationComputeResource",
    "MaterializationSettings",
    "MedianStoppingPolicy",
    "ModelContainer",
    "ModelContainerProperties",
    "ModelContainerResourceArmPaginatedResult",
    "ModelVersion",
    "ModelVersionProperties",
    "ModelVersionResourceArmPaginatedResult",
    "MonitorComputeConfigurationBase",
    "MonitorComputeIdentityBase",
    "MonitorDefinition",
    "MonitorEmailNotificationSettings",
    "MonitorNotificationSettings",
    "MonitorServerlessSparkCompute",
    "MonitoringFeatureFilterBase",
    "MonitoringInputDataBase",
    "MonitoringSignalBase",
    "MonitoringTarget",
    "MonitoringThreshold",
    "Mpi",
    "NCrossValidations",
    "NlpVertical",
    "NlpVerticalFeaturizationSettings",
    "NlpVerticalLimitSettings",
    "NodeStateCounts",
    "Nodes",
    "NoneAuthTypeWorkspaceConnectionProperties",
    "NoneDatastoreCredentials",
    "NotebookAccessTokenResult",
    "NotebookPreparationError",
    "NotebookResourceInfo",
    "NotificationSetting",
    "NumericalDataDriftMetricThreshold",
    "NumericalDataQualityMetricThreshold",
    "NumericalPredictionDriftMetricThreshold",
    "Objective",
    "OnlineDeployment",
    "OnlineDeploymentProperties",
    "OnlineDeploymentTrackedResourceArmPaginatedResult",
    "OnlineEndpoint",
    "OnlineEndpointProperties",
    "OnlineEndpointTrackedResourceArmPaginatedResult",
    "OnlineRequestSettings",
    "OnlineScaleSettings",
    "Operation",
    "OperationDisplay",
    "OperationListResult",
    "OutboundRule",
    "OutboundRuleBasicResource",
    "OutboundRuleListResult",
    "OutputPathAssetReference",
    "PATAuthTypeWorkspaceConnectionProperties",
    "PaginatedComputeResourcesList",
    "PartialBatchDeployment",
    "PartialBatchDeploymentPartialMinimalTrackedResourceWithProperties",
    "PartialManagedServiceIdentity",
    "PartialMinimalTrackedResource",
    "PartialMinimalTrackedResourceWithIdentity",
    "PartialMinimalTrackedResourceWithSku",
    "PartialRegistryPartialTrackedResource",
    "PartialSku",
    "Password",
    "PendingUploadCredentialDto",
    "PendingUploadRequestDto",
    "PendingUploadResponseDto",
    "PersonalComputeInstanceSettings",
    "PipelineJob",
    "PredictionDriftMetricThresholdBase",
    "PredictionDriftMonitoringSignal",
    "PrivateEndpoint",
    "PrivateEndpointConnection",
    "PrivateEndpointConnectionListResult",
    "PrivateEndpointDestination",
    "PrivateEndpointOutboundRule",
    "PrivateEndpointResource",
    "PrivateLinkResource",
    "PrivateLinkResourceListResult",
    "PrivateLinkServiceConnectionState",
    "ProbeSettings",
    "ProxyResource",
    "PyTorch",
    "QueueSettings",
    "QuotaBaseProperties",
    "QuotaUpdateParameters",
    "RandomSamplingAlgorithm",
    "Recurrence",
    "RecurrenceSchedule",
    "RecurrenceTrigger",
    "RegenerateEndpointKeysRequest",
    "Registry",
    "RegistryListCredentialsResult",
    "RegistryPartialManagedServiceIdentity",
    "RegistryPrivateEndpointConnection",
    "RegistryPrivateLinkServiceConnectionState",
    "RegistryRegionArmDetails",
    "RegistryTrackedResourceArmPaginatedResult",
    "Regression",
    "RegressionTrainingSettings",
    "Resource",
    "ResourceBase",
    "ResourceConfiguration",
    "ResourceId",
    "ResourceName",
    "ResourceQuota",
    "RollingInputData",
    "Route",
    "SASAuthTypeWorkspaceConnectionProperties",
    "SASCredential",
    "SASCredentialDto",
    "SamplingAlgorithm",
    "SasDatastoreCredentials",
    "SasDatastoreSecrets",
    "ScaleSettings",
    "ScaleSettingsInformation",
    "Schedule",
    "ScheduleActionBase",
    "ScheduleBase",
    "ScheduleProperties",
    "ScheduleResourceArmPaginatedResult",
    "ScriptReference",
    "ScriptsToExecute",
    "Seasonality",
    "ServerlessComputeSettings",
    "ServiceManagedResourcesSettings",
    "ServicePrincipalDatastoreCredentials",
    "ServicePrincipalDatastoreSecrets",
    "ServiceTagDestination",
    "ServiceTagOutboundRule",
    "SetupScripts",
    "SharedPrivateLinkResource",
    "Sku",
    "SkuCapacity",
    "SkuResource",
    "SkuResourceArmPaginatedResult",
    "SkuSetting",
    "SslConfiguration",
    "StackEnsembleSettings",
    "StaticInputData",
    "StorageAccountDetails",
    "SweepJob",
    "SweepJobLimits",
    "SynapseSpark",
    "SynapseSparkProperties",
    "SystemCreatedAcrAccount",
    "SystemCreatedStorageAccount",
    "SystemData",
    "SystemService",
    "TableVertical",
    "TableVerticalFeaturizationSettings",
    "TableVerticalLimitSettings",
    "TargetLags",
    "TargetRollingWindowSize",
    "TargetUtilizationScaleSettings",
    "TensorFlow",
    "TextClassification",
    "TextClassificationMultilabel",
    "TextNer",
    "TmpfsOptions",
    "TopNFeaturesByAttribution",
    "TrackedResource",
    "TrainingSettings",
    "TrialComponent",
    "TriggerBase",
    "TritonModelJobInput",
    "TritonModelJobOutput",
    "TruncationSelectionPolicy",
    "UpdateWorkspaceQuotas",
    "UpdateWorkspaceQuotasResult",
    "UriFileDataVersion",
    "UriFileJobInput",
    "UriFileJobOutput",
    "UriFolderDataVersion",
    "UriFolderJobInput",
    "UriFolderJobOutput",
    "Usage",
    "UsageName",
    "UserAccountCredentials",
    "UserAssignedIdentity",
    "UserCreatedAcrAccount",
    "UserCreatedStorageAccount",
    "UserIdentity",
    "UsernamePasswordAuthTypeWorkspaceConnectionProperties",
    "VirtualMachine",
    "VirtualMachineImage",
    "VirtualMachineSchema",
    "VirtualMachineSchemaProperties",
    "VirtualMachineSecrets",
    "VirtualMachineSecretsSchema",
    "VirtualMachineSize",
    "VirtualMachineSizeListResult",
    "VirtualMachineSshCredentials",
    "VolumeDefinition",
    "VolumeOptions",
    "Webhook",
    "Workspace",
    "WorkspaceConnectionManagedIdentity",
    "WorkspaceConnectionPersonalAccessToken",
    "WorkspaceConnectionPropertiesV2",
    "WorkspaceConnectionPropertiesV2BasicResource",
    "WorkspaceConnectionPropertiesV2BasicResourceArmPaginatedResult",
    "WorkspaceConnectionSharedAccessSignature",
    "WorkspaceConnectionUsernamePassword",
    "WorkspaceListResult",
    "WorkspaceUpdateParameters",
    "ActionType",
    "AllocationState",
    "ApplicationSharingPolicy",
    "AssetProvisioningState",
    "AutoRebuildSetting",
    "Autosave",
    "BatchLoggingLevel",
    "BatchOutputAction",
    "BillingCurrency",
    "BlockedTransformers",
    "Caching",
    "CategoricalDataDriftMetric",
    "CategoricalDataQualityMetric",
    "CategoricalPredictionDriftMetric",
    "ClassificationModels",
    "ClassificationMultilabelPrimaryMetrics",
    "ClassificationPrimaryMetrics",
    "ClusterPurpose",
    "ComputeInstanceAuthorizationType",
    "ComputeInstanceState",
    "ComputePowerAction",
    "ComputeRecurrenceFrequency",
    "ComputeTriggerType",
    "ComputeType",
    "ComputeWeekDay",
    "ConnectionAuthType",
    "ConnectionCategory",
    "ContainerType",
    "CreatedByType",
    "CredentialsType",
    "DataAvailabilityStatus",
    "DataReferenceCredentialType",
    "DataType",
    "DatastoreType",
    "DeploymentProvisioningState",
    "DiagnoseResultLevel",
    "DistributionType",
    "EarlyTerminationPolicyType",
    "EgressPublicNetworkAccessType",
    "EmailNotificationEnableType",
    "EncryptionStatus",
    "EndpointAuthMode",
    "EndpointComputeType",
    "EndpointProvisioningState",
    "EndpointServiceConnectionStatus",
    "EnvironmentType",
    "EnvironmentVariableType",
    "FeatureAttributionMetric",
    "FeatureDataType",
    "FeatureImportanceMode",
    "FeatureLags",
    "FeaturizationMode",
    "ForecastHorizonMode",
    "ForecastingModels",
    "ForecastingPrimaryMetrics",
    "Goal",
    "IdentityConfigurationType",
    "ImageType",
    "InputDeliveryMode",
    "InstanceSegmentationPrimaryMetrics",
    "IsolationMode",
    "JobInputType",
    "JobLimitsType",
    "JobOutputType",
    "JobStatus",
    "JobTier",
    "JobType",
    "KeyType",
    "LearningRateScheduler",
    "ListViewType",
    "LoadBalancerType",
    "LogVerbosity",
    "ManagedNetworkStatus",
    "ManagedServiceIdentityType",
    "MaterializationStoreType",
    "ModelSize",
    "ModelTaskType",
    "MonitorComputeIdentityType",
    "MonitorComputeType",
    "MonitoringFeatureDataType",
    "MonitoringFeatureFilterType",
    "MonitoringInputDataType",
    "MonitoringNotificationType",
    "MonitoringSignalType",
    "MountAction",
    "MountState",
    "NCrossValidationsMode",
    "Network",
    "NodeState",
    "NodesValueType",
    "NumericalDataDriftMetric",
    "NumericalDataQualityMetric",
    "NumericalPredictionDriftMetric",
    "ObjectDetectionPrimaryMetrics",
    "OperatingSystemType",
    "OperationName",
    "OperationStatus",
    "OperationTrigger",
    "OrderString",
    "Origin",
    "OsType",
    "OutputDeliveryMode",
    "PendingUploadCredentialType",
    "PendingUploadType",
    "PrivateEndpointConnectionProvisioningState",
    "PrivateEndpointServiceConnectionStatus",
    "Protocol",
    "ProvisioningState",
    "ProvisioningStatus",
    "PublicNetworkAccess",
    "PublicNetworkAccessType",
    "QuotaUnit",
    "RandomSamplingAlgorithmRule",
    "RecurrenceFrequency",
    "ReferenceType",
    "RegressionModels",
    "RegressionPrimaryMetrics",
    "RemoteLoginPortPublicAccess",
    "RuleAction",
    "RuleCategory",
    "RuleStatus",
    "RuleType",
    "SamplingAlgorithmType",
    "ScaleType",
    "ScheduleActionType",
    "ScheduleListViewType",
    "ScheduleProvisioningState",
    "ScheduleProvisioningStatus",
    "ScheduleStatus",
    "SeasonalityMode",
    "SecretsType",
    "ServiceDataAccessAuthIdentity",
    "ShortSeriesHandlingConfiguration",
    "SkuScaleType",
    "SkuTier",
    "SourceType",
    "SshPublicAccess",
    "SslConfigStatus",
    "StackMetaLearnerType",
    "Status",
    "StochasticOptimizer",
    "StorageAccountType",
    "TargetAggregationFunction",
    "TargetLagsMode",
    "TargetRollingWindowSizeMode",
    "TaskType",
    "TriggerType",
    "UnderlyingResourceAction",
    "UnitOfMeasure",
    "UsageUnit",
    "UseStl",
    "VMPriceOSType",
    "VMTier",
    "ValidationMetricType",
    "ValueFormat",
    "VmPriority",
    "VolumeDefinitionType",
    "WebhookType",
    "WeekDay",
]
__all__.extend([p for p in _patch_all if p not in __all__])
_patch_sdk()



from pathlib import Path

import pydash
import pytest
from test_utilities.utils import parse_local_path

from azure.ai.ml import load_component
from azure.ai.ml._utils.utils import load_yaml
from azure.ai.ml.entities import Component
from azure.ai.ml.entities._component.parallel_component import ParallelComponent
from azure.ai.ml.entities._job.pipeline._io import PipelineInput

from .._util import _COMPONENT_TIMEOUT_SECOND


@pytest.mark.timeout(_COMPONENT_TIMEOUT_SECOND)
@pytest.mark.unittest
@pytest.mark.pipeline_test
class TestParallelComponentEntity:
    def test_component_load(self):
        # code is specified in yaml, value is respected
        component_yaml = "./tests/test_configs/components/basic_parallel_component_score.yml"
        parallel_component = load_component(
            source=component_yaml,
        )
        assert parallel_component.task.get("type") == "run_function"
        assert parallel_component.code == "../python"

    def test_command_component_to_dict(self):
        # Test optional params exists in component dict
        yaml_path = "./tests/test_configs/components/basic_parallel_component_score.yml"
        yaml_dict = load_yaml(yaml_path)
        yaml_dict["mock_option_param"] = {"mock_key": "mock_val"}
        parallel_component = ParallelComponent._load(data=yaml_dict, yaml_path=yaml_path)
        assert parallel_component._other_parameter.get("mock_option_param") == yaml_dict["mock_option_param"]

    def test_parallel_component_entity(self):
        task = {
            "type": "run_function",
            "code": "../python",
            "entry_script": "pass_through.py",
            "program_arguments": "--label ${{inputs.label}} --model ${{inputs.score_model}} --output_scored_result ${{outputs.scored_result}}",
            "append_row_to": "${{outputs.scoring_summary}}",
            "environment": "AzureML-Minimal:2",
        }

        component = ParallelComponent(
            name="batch_score",
            display_name="BatchScore",
            description="parallel component for batch score",
            version="1.0.0",
            outputs={
                "scored_result": {"type": "mltable"},
                "scoring_summary": {"type": "uri_file"},
            },
            resources={"instance_count": 2},
            retry_settings={"max_retries": 10, "timeout": 3},
            mini_batch_size="10kb",
            input_data="${{inputs.score_input}}",
            max_concurrency_per_instance=12,
            error_threshold=10,
            mini_batch_error_threshold=5,
            logging_level="INFO",
            task=task,
            base_path="./tests/test_configs/components",
        )
        omit_fields = [
            "properties.component_spec.$schema",
            "properties.component_spec.inputs",
            "properties.component_spec._source",
            "properties.properties.client_component_hash",
        ]
        component_dict = component._to_rest_object().as_dict()
        component_dict = pydash.omit(component_dict, *omit_fields)

        yaml_path = "./tests/test_configs/components/basic_parallel_component_score.yml"
        yaml_component = load_component(source=yaml_path)
        yaml_component_dict = yaml_component._to_rest_object().as_dict()
        yaml_component_dict = pydash.omit(yaml_component_dict, *omit_fields)

        assert component_dict == yaml_component_dict

    def test_parallel_component_version_as_a_function_with_inputs(self):
        yaml_path = "./tests/test_configs/components/helloworld_parallel.yml"
        yaml_component_version = load_component(source=yaml_path)
        expected_rest_component = {
            "componentId": "fake_component",
            "_source": "YAML.COMPONENT",
            "input_data": "${{inputs.component_in_path}}",
            "inputs": {
                "model": {"job_input_type": "literal", "value": "SVM"},
                "label": {"job_input_type": "literal", "value": "test"},
                "component_in_path": {
                    "job_input_type": "literal",
                    "value": "${{parent.inputs.pipeline_input}}",
                },
            },
            "input_data": "${{inputs.component_in_path}}",
            "type": "parallel",
            "mini_batch_size": 10485760,
            "task": {
                "append_row_to": "${{outputs.scoring_summary}}",
                "program_arguments": "--label ${{inputs.label}} --model ${{inputs.model}}",
                "code": parse_local_path("../python", yaml_component_version.base_path),
                "entry_script": "score.py",
                "environment": "azureml:AzureML-sklearn-1.0-ubuntu20.04-py38-cpu:33",
                "type": "run_function",
            },
        }
        pipeline_input = PipelineInput(name="pipeline_input", owner="pipeline", meta=None)
        yaml_component = yaml_component_version(model="SVM", label="test", component_in_path=pipeline_input)

        yaml_component._component = "fake_component"
        rest_yaml_component = yaml_component._to_rest_object()

        assert rest_yaml_component == expected_rest_component

    def test_parallel_component_run_settings_picked_up(self):
        yaml_path = "./tests/test_configs/components/parallel_component_with_run_settings.yml"
        parallel_component = load_component(source=yaml_path)
        parallel_node = parallel_component()
        # Normally, during initiation of nodes, the settings from the yaml file shouldn't be changed
        assert parallel_component.resources.instance_count == parallel_node.resources.instance_count == 1
        assert parallel_component.max_concurrency_per_instance == parallel_node.max_concurrency_per_instance == 16
        assert parallel_component.retry_settings == parallel_node.retry_settings
        assert parallel_component.retry_settings.timeout == 12345



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import copy
import json
import os
from typing import Tuple

import pytest
from devtools_testutils import AzureRecordedTestCase, is_live
from test_utilities.utils import assert_final_job_status, get_automl_job_properties

from azure.ai.ml import MLClient, automl
from azure.ai.ml.constants._common import AssetTypes
from azure.ai.ml.entities import Data
from azure.ai.ml.entities._inputs_outputs import Input
from azure.ai.ml.entities._job.automl import SearchSpace
from azure.ai.ml.entities._job.automl.image import ImageClassificationMultilabelJob, ImageClassificationSearchSpace
from azure.ai.ml.operations._run_history_constants import JobStatus
from azure.ai.ml.sweep import BanditPolicy, Choice, Uniform


@pytest.mark.automl_test
@pytest.mark.usefixtures(
    "recorded_test",
    "mock_asset_name",
    "mock_code_hash",
)
@pytest.mark.skipif(condition=not is_live(), reason="Datasets downloaded by test are too large to record reliably")
class TestAutoMLImageClassificationMultilabel(AzureRecordedTestCase):
    def _create_jsonl_multilabel(self, client: MLClient, train_path: str, val_path: str):
        src_images = "./multilabelFridgeObjects/"

        train_validation_ratio = 5

        # Path to the training and validation files
        train_annotations_file = os.path.join(train_path, "train_annotations.jsonl")
        validation_annotations_file = os.path.join(val_path, "validation_annotations.jsonl")

        multilabel_data = Data(path="./multilabelFridgeObjects", type=AssetTypes.URI_FOLDER)
        uri_folder_data_asset = client.data.create_or_update(multilabel_data)

        # Baseline of json line dictionary
        json_line_sample = {
            "image_url": uri_folder_data_asset.path,
            "label": [],
        }

        # Path to the labels file.
        labelFile = os.path.join(src_images, "labels.csv")

        # Read each annotation and convert it to jsonl line
        with open(train_annotations_file, "w") as train_f:
            with open(validation_annotations_file, "w") as validation_f:
                with open(labelFile, "r") as labels:
                    for i, line in enumerate(labels):
                        # Skipping the title line and any empty lines.
                        if i == 0 or len(line.strip()) == 0:
                            continue
                        line_split = line.strip().split(",")
                        if len(line_split) != 2:
                            print("Skipping the invalid line: {}".format(line))
                            continue
                        json_line = dict(json_line_sample)
                        json_line["image_url"] += f"images/{line_split[0]}"
                        json_line["label"] = line_split[1].strip().split(" ")

                        if i % train_validation_ratio == 0:
                            # validation annotation
                            validation_f.write(json.dumps(json_line) + "\n")
                        else:
                            # train annotation
                            train_f.write(json.dumps(json_line) + "\n")

    @pytest.mark.parametrize("components", [(False), (True)])
    def test_image_classification_multilabel_run(
        self,
        image_classification_multilabel_dataset: Tuple[Input, Input],
        client: MLClient,
        components: bool,
    ) -> None:
        # Note: this test launches two jobs in order to avoid calling the dataset fixture more than once. Ideally, it
        # would have sufficed to mark the fixture with session scope, but pytest-xdist breaks this functionality:
        # https://github.com/pytest-dev/pytest-xdist/issues/271.

        # Get training and validation data paths
        train_path, val_path = image_classification_multilabel_dataset

        # Create jsonl file
        self._create_jsonl_multilabel(client=client, train_path=train_path, val_path=val_path)

        training_data = Input(type=AssetTypes.MLTABLE, path=train_path)
        validation_data = Input(type=AssetTypes.MLTABLE, path=val_path)

        # Make generic multilabel classification job
        image_classification_multilabel_job = automl.image_classification_multilabel(
            compute="gpu-cluster",
            experiment_name="image-e2e-tests",
            training_data=training_data,
            validation_data=validation_data,
            target_column_name="label",
            primary_metric="iou",
            # These are temporal properties needed in Private Preview
            properties=get_automl_job_properties(),
        )

        # Configure sweep job
        image_classification_multilabel_job_sweep = copy.deepcopy(image_classification_multilabel_job)
        image_classification_multilabel_job_sweep.set_training_parameters(early_stopping=True, evaluation_frequency=1)
        image_classification_multilabel_job_sweep.set_limits(max_trials=2, max_concurrent_trials=2)

        if components:
            # Configure components sweep job search space
            image_classification_multilabel_job_sweep.extend_search_space(
                [
                    SearchSpace(
                        model_name=Choice(["microsoft/beit-base-patch16-224"]),
                        number_of_epochs=Choice([1]),
                        gradient_accumulation_step=Choice([1]),
                        learning_rate=Choice([0.005]),
                    ),
                    SearchSpace(
                        model_name=Choice(["seresnext"]),
                        # model-specific, valid_resize_size should be larger or equal than valid_crop_size
                        validation_resize_size=Choice([288]),
                        validation_crop_size=Choice([224]),  # model-specific
                        training_crop_size=Choice([224]),  # model-specific
                        number_of_epochs=Choice([1]),
                    ),
                ]
            )
            image_classification_multilabel_job_sweep.set_sweep(
                sampling_algorithm="Grid",
                early_termination=BanditPolicy(evaluation_interval=2, slack_factor=0.2, delay_evaluation=6),
            )

            image_classification_multilabel_job_individual = copy.deepcopy(image_classification_multilabel_job)
            image_classification_multilabel_job_individual.set_training_parameters(
                model_name="microsoft/beit-base-patch16-224", number_of_epochs=1
            )
            image_classification_multilabel_job_reuse = copy.deepcopy(image_classification_multilabel_job_individual)
        else:
            # Configure runtime sweep job search space
            image_classification_multilabel_job_sweep.extend_search_space(
                [
                    SearchSpace(
                        model_name=Choice(["vitb16r224"]),
                        learning_rate=Uniform(0.005, 0.05),
                        number_of_epochs=Choice([1]),
                        gradient_accumulation_step=Choice([1, 2]),
                    ),
                    SearchSpace(
                        model_name=Choice(["seresnext"]),
                        learning_rate=Uniform(0.005, 0.05),
                        # model-specific, valid_resize_size should be larger or equal than valid_crop_size
                        validation_resize_size=Choice([288, 320, 352]),
                        validation_crop_size=Choice([224, 256]),  # model-specific
                        training_crop_size=Choice([224, 256]),  # model-specific
                        number_of_epochs=Choice([1]),
                    ),
                ]
            )
            image_classification_multilabel_job_sweep.set_sweep(
                sampling_algorithm="Random",
                early_termination=BanditPolicy(evaluation_interval=2, slack_factor=0.2, delay_evaluation=6),
            )

            # Configure AutoMode job
            image_classification_multilabel_job_automode = copy.deepcopy(image_classification_multilabel_job)
            image_classification_multilabel_job_automode.set_limits(max_trials=2, max_concurrent_trials=2)

        # Trigger sweep job and then AutoMode job
        submitted_job_sweep = client.jobs.create_or_update(image_classification_multilabel_job_sweep)
        if components:
            submitted_job_individual_components = client.jobs.create_or_update(
                image_classification_multilabel_job_individual
            )
            submitted_job_components_reuse = client.jobs.create_or_update(image_classification_multilabel_job_reuse)
        else:
            submitted_job_automode = client.jobs.create_or_update(image_classification_multilabel_job_automode)

        # Assert completion of sweep job
        assert_final_job_status(
            submitted_job_sweep, client, ImageClassificationMultilabelJob, JobStatus.COMPLETED, deadline=3600
        )

        if components:
            assert_final_job_status(
                submitted_job_individual_components,
                client,
                ImageClassificationMultilabelJob,
                JobStatus.COMPLETED,
                deadline=3600,
            )
            assert_final_job_status(
                submitted_job_components_reuse,
                client,
                ImageClassificationMultilabelJob,
                JobStatus.COMPLETED,
                deadline=3600,
            )
        else:
            # Assert completion of Automode job
            assert_final_job_status(
                submitted_job_automode, client, ImageClassificationMultilabelJob, JobStatus.COMPLETED, deadline=3600
            )



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
from typing import Any, Dict, List, Optional, Type, Union

from azure.ai.ml.exceptions import ErrorCategory, ErrorTarget, ValidationErrorType, ValidationException

from ..job_limits import SweepJobLimits
from ..job_resource_configuration import JobResourceConfiguration
from ..queue_settings import QueueSettings
from .early_termination_policy import (
    BanditPolicy,
    EarlyTerminationPolicy,
    EarlyTerminationPolicyType,
    MedianStoppingPolicy,
    TruncationSelectionPolicy,
)
from .objective import Objective
from .sampling_algorithm import (
    BayesianSamplingAlgorithm,
    GridSamplingAlgorithm,
    RandomSamplingAlgorithm,
    RestBayesianSamplingAlgorithm,
    RestGridSamplingAlgorithm,
    RestRandomSamplingAlgorithm,
    RestSamplingAlgorithm,
    SamplingAlgorithm,
    SamplingAlgorithmType,
)

SAMPLING_ALGORITHM_TO_REST_CONSTRUCTOR: Dict[SamplingAlgorithmType, Type[RestSamplingAlgorithm]] = {
    SamplingAlgorithmType.RANDOM: RestRandomSamplingAlgorithm,
    SamplingAlgorithmType.GRID: RestGridSamplingAlgorithm,
    SamplingAlgorithmType.BAYESIAN: RestBayesianSamplingAlgorithm,
}

SAMPLING_ALGORITHM_CONSTRUCTOR: Dict[SamplingAlgorithmType, Type[SamplingAlgorithm]] = {
    SamplingAlgorithmType.RANDOM: RandomSamplingAlgorithm,
    SamplingAlgorithmType.GRID: GridSamplingAlgorithm,
    SamplingAlgorithmType.BAYESIAN: BayesianSamplingAlgorithm,
}


class ParameterizedSweep:  # pylint:disable=too-many-instance-attributes
    """Shared logic for standalone and pipeline sweep job."""

    def __init__(
        self,
        limits: Optional[SweepJobLimits] = None,
        sampling_algorithm: Optional[Union[str, SamplingAlgorithm]] = None,
        objective: Optional[Union[Dict, Objective]] = None,
        early_termination: Optional[Any] = None,
        search_space: Optional[Dict] = None,
        queue_settings: Optional[QueueSettings] = None,
        resources: Optional[Union[dict, JobResourceConfiguration]] = None,
    ) -> None:
        """
        :param limits: Limits for sweep job.
        :type limits: ~azure.ai.ml.sweep.SweepJobLimits
        :param sampling_algorithm: Sampling algorithm for sweep job.
        :type sampling_algorithm: ~azure.ai.ml.sweep.SamplingAlgorithm
        :param objective: Objective for sweep job.
        :type objective: ~azure.ai.ml.sweep.Objective
        :param early_termination: Early termination policy for sweep job.
        :type early_termination: ~azure.ai.ml.entities._job.sweep.early_termination_policy.EarlyTerminationPolicy
        :param search_space: Search space for sweep job.
        :type search_space: Dict[str, Union[
            ~azure.ai.ml.sweep.Choice,
            ~azure.ai.ml.sweep.LogNormal,
            ~azure.ai.ml.sweep.LogUniform,
            ~azure.ai.ml.sweep.Normal,
            ~azure.ai.ml.sweep.QLogNormal,
            ~azure.ai.ml.sweep.QLogUniform,
            ~azure.ai.ml.sweep.QNormal,
            ~azure.ai.ml.sweep.QUniform,
            ~azure.ai.ml.sweep.Randint,
            ~azure.ai.ml.sweep.Uniform

            ]]
        :param queue_settings: Queue settings for sweep job.
        :type queue_settings: ~azure.ai.ml.entities.QueueSettings
        :param resources: Compute Resource configuration for the job.
        :type resources: ~azure.ai.ml.entities.ResourceConfiguration
        """
        self.sampling_algorithm = sampling_algorithm
        self.early_termination = early_termination  # type: ignore[assignment]
        self._limits = limits
        self.search_space = search_space
        self.queue_settings = queue_settings
        self.objective: Optional[Objective] = None
        self.resources = resources

        if isinstance(objective, Dict):
            self.objective = Objective(**objective)
        else:
            self.objective = objective

    @property
    def resources(self) -> Optional[Union[dict, JobResourceConfiguration]]:
        """Resources for sweep job.

        :returns: Resources for sweep job.
        :rtype: ~azure.ai.ml.entities.ResourceConfiguration
        """
        return self._resources

    @resources.setter
    def resources(self, value: Optional[Union[dict, JobResourceConfiguration]]) -> None:
        """Set Resources for sweep job.

        :param value: Compute Resource configuration for the job.
        :type value: ~azure.ai.ml.entities.ResourceConfiguration
        """
        if isinstance(value, dict):
            value = JobResourceConfiguration(**value)
        self._resources = value

    @property
    def limits(self) -> Optional[SweepJobLimits]:
        """Limits for sweep job.

        :returns: Limits for sweep job.
        :rtype: ~azure.ai.ml.sweep.SweepJobLimits
        """
        return self._limits

    @limits.setter
    def limits(self, value: SweepJobLimits) -> None:
        """Set limits for sweep job.

        :param value: Limits for sweep job.
        :type value: ~azure.ai.ml.sweep.SweepJobLimits
        """
        if not isinstance(value, SweepJobLimits):
            msg = f"limits must be SweepJobLimits but get {type(value)} instead"
            raise ValidationException(
                message=msg,
                no_personal_data_message=msg,
                target=ErrorTarget.SWEEP_JOB,
                error_category=ErrorCategory.USER_ERROR,
                error_type=ValidationErrorType.INVALID_VALUE,
            )
        self._limits = value

    def set_resources(
        self,
        *,
        instance_type: Optional[Union[str, List[str]]] = None,
        instance_count: Optional[int] = None,
        locations: Optional[List[str]] = None,
        properties: Optional[Dict] = None,
        docker_args: Optional[str] = None,
        shm_size: Optional[str] = None,
    ) -> None:
        """Set resources for Sweep.

        :keyword instance_type: The instance type to use for the job.
        :paramtype instance_type: Optional[Union[str, List[str]]]
        :keyword instance_count: The number of instances to use for the job.
        :paramtype instance_count: Optional[int]
        :keyword locations: The locations to use for the job.
        :paramtype locations: Optional[List[str]]
        :keyword properties: The properties for the job.
        :paramtype properties: Optional[Dict]
        :keyword docker_args: The docker arguments for the job.
        :paramtype docker_args: Optional[str]
        :keyword shm_size: The shared memory size for the job.
        :paramtype shm_size: Optional[str]
        """
        if self.resources is None:
            self.resources = JobResourceConfiguration()

        if not isinstance(self.resources, dict):
            if locations is not None:
                self.resources.locations = locations
            if instance_type is not None:
                self.resources.instance_type = instance_type
            if instance_count is not None:
                self.resources.instance_count = instance_count
            if properties is not None:
                self.resources.properties = properties
            if docker_args is not None:
                self.resources.docker_args = docker_args
            if shm_size is not None:
                self.resources.shm_size = shm_size

    def set_limits(
        self,
        *,
        max_concurrent_trials: Optional[int] = None,
        max_total_trials: Optional[int] = None,
        timeout: Optional[int] = None,
        trial_timeout: Optional[int] = None,
    ) -> None:
        """Set limits for Sweep node. Leave parameters as None if you don't want to update corresponding values.

        :keyword max_concurrent_trials: maximum concurrent trial number.
        :paramtype max_concurrent_trials: int
        :keyword max_total_trials: maximum total trial number.
        :paramtype max_total_trials: int
        :keyword timeout: total timeout in seconds for sweep node
        :paramtype timeout: int
        :keyword trial_timeout: timeout in seconds for each trial
        :paramtype trial_timeout: int
        """
        # Looks related to https://github.com/pylint-dev/pylint/issues/3502, still an open issue
        # pylint:disable=attribute-defined-outside-init
        if self._limits is None:
            self._limits = SweepJobLimits(
                max_concurrent_trials=max_concurrent_trials,
                max_total_trials=max_total_trials,
                timeout=timeout,
                trial_timeout=trial_timeout,
            )
        else:
            if self.limits is not None:
                if max_concurrent_trials is not None:
                    self.limits.max_concurrent_trials = max_concurrent_trials
                if max_total_trials is not None:
                    self.limits.max_total_trials = max_total_trials
                if timeout is not None:
                    self.limits.timeout = timeout
                if trial_timeout is not None:
                    self.limits.trial_timeout = trial_timeout

    def set_objective(self, *, goal: Optional[str] = None, primary_metric: Optional[str] = None) -> None:
        """Set the sweep object.. Leave parameters as None if you don't want to update corresponding values.

        :keyword goal: Defines supported metric goals for hyperparameter tuning. Acceptable values are:
            "minimize" and "maximize".
        :paramtype goal: str
        :keyword primary_metric: Name of the metric to optimize.
        :paramtype primary_metric: str
        """

        if self.objective is not None:
            if goal:
                self.objective.goal = goal
            if primary_metric:
                self.objective.primary_metric = primary_metric
        else:
            self.objective = Objective(goal=goal, primary_metric=primary_metric)

    @property
    def sampling_algorithm(self) -> Optional[Union[str, SamplingAlgorithm]]:
        """Sampling algorithm for sweep job.

        :returns: Sampling algorithm for sweep job.
        :rtype: ~azure.ai.ml.sweep.SamplingAlgorithm
        """
        return self._sampling_algorithm

    @sampling_algorithm.setter
    def sampling_algorithm(self, value: Optional[Union[SamplingAlgorithm, str]] = None) -> None:
        """Set sampling algorithm for sweep job.

        :param value: Sampling algorithm for sweep job.
        :type value: ~azure.ai.ml.sweep.SamplingAlgorithm
        """
        if value is None:
            self._sampling_algorithm = None
        elif isinstance(value, SamplingAlgorithm) or (
            isinstance(value, str) and value.lower().capitalize() in SAMPLING_ALGORITHM_CONSTRUCTOR
        ):
            self._sampling_algorithm = value
        else:
            msg = f"unsupported sampling algorithm: {value}"
            raise ValidationException(
                message=msg,
                no_personal_data_message=msg,
                target=ErrorTarget.SWEEP_JOB,
                error_category=ErrorCategory.USER_ERROR,
                error_type=ValidationErrorType.INVALID_VALUE,
            )

    def _get_rest_sampling_algorithm(self) -> RestSamplingAlgorithm:
        # TODO: self.sampling_algorithm will always return SamplingAlgorithm
        if isinstance(self.sampling_algorithm, SamplingAlgorithm):
            return self.sampling_algorithm._to_rest_object()  # pylint: disable=protected-access

        if isinstance(self.sampling_algorithm, str):
            return SAMPLING_ALGORITHM_CONSTRUCTOR[  # pylint: disable=protected-access
                SamplingAlgorithmType(self.sampling_algorithm.lower().capitalize())
            ]()._to_rest_object()

        msg = f"Received unsupported value {self._sampling_algorithm} as the sampling algorithm"
        raise ValidationException(
            message=msg,
            no_personal_data_message=msg,
            target=ErrorTarget.SWEEP_JOB,
            error_category=ErrorCategory.USER_ERROR,
            error_type=ValidationErrorType.INVALID_VALUE,
        )

    @property
    def early_termination(self) -> Optional[Union[str, EarlyTerminationPolicy]]:
        """Early termination policy for sweep job.

        :returns: Early termination policy for sweep job.
        :rtype: ~azure.ai.ml.entities._job.sweep.early_termination_policy.EarlyTerminationPolicy
        """
        return self._early_termination

    @early_termination.setter
    def early_termination(self, value: Any) -> None:
        """Set early termination policy for sweep job.

        :param value: Early termination policy for sweep job.
        :type value: ~azure.ai.ml.entities._job.sweep.early_termination_policy.EarlyTerminationPolicy
        """
        self._early_termination: Optional[Union[str, EarlyTerminationPolicy]]
        if value is None:
            self._early_termination = None
        elif isinstance(value, EarlyTerminationPolicy):
            self._early_termination = value
        elif isinstance(value, str):
            value = value.lower().capitalize()
            if value == EarlyTerminationPolicyType.BANDIT:
                self._early_termination = BanditPolicy()
            elif value == EarlyTerminationPolicyType.MEDIAN_STOPPING:
                self._early_termination = MedianStoppingPolicy()
            elif value == EarlyTerminationPolicyType.TRUNCATION_SELECTION:
                self._early_termination = TruncationSelectionPolicy()
            else:
                msg = f"Received unsupported value {value} as the early termination policy"
                raise ValidationException(
                    message=msg,
                    no_personal_data_message=msg,
                    target=ErrorTarget.SWEEP_JOB,
                    error_category=ErrorCategory.USER_ERROR,
                    error_type=ValidationErrorType.INVALID_VALUE,
                )
        else:
            msg = f"Received unsupported value of type {type(value)} as the early termination policy"
            raise ValidationException(
                message=msg,
                no_personal_data_message=msg,
                target=ErrorTarget.SWEEP_JOB,
                error_category=ErrorCategory.USER_ERROR,
                error_type=ValidationErrorType.INVALID_VALUE,
            )



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
import json
import os
from typing import Any, Dict, Optional, Union

from azure.ai.ml import Input, Output
from azure.ai.ml._schema import PathAwareSchema
from azure.ai.ml._schema.pipeline.control_flow_job import ParallelForSchema
from azure.ai.ml._utils.utils import is_data_binding_expression
from azure.ai.ml.constants import AssetTypes
from azure.ai.ml.constants._component import ComponentParameterTypes, ControlFlowType
from azure.ai.ml.entities import Component, Pipeline
from azure.ai.ml.entities._builders import BaseNode
from azure.ai.ml.entities._builders.control_flow_node import LoopNode
from azure.ai.ml.entities._job.pipeline._io import NodeOutput, PipelineInput
from azure.ai.ml.entities._job.pipeline._io.mixin import NodeIOMixin
from azure.ai.ml.entities._util import convert_ordered_dict_to_dict, validate_attribute_type
from azure.ai.ml.entities._validation import MutableValidationResult
from azure.ai.ml.exceptions import UserErrorException


class ParallelFor(LoopNode, NodeIOMixin):
    """Parallel for loop node in the pipeline job. By specifying the loop body and aggregated items, a job-level
    parallel for loop can be implemented. It will be initialized when calling dsl.parallel_for or when loading the
    pipeline yml containing parallel_for node. Please do not manually initialize this class.

    :param body: Pipeline job for the parallel for loop body.
    :type body: ~azure.ai.ml.entities.Pipeline
    :param items: The loop body's input which will bind to the loop node.
    :type items: typing.Union[list, dict, str, ~azure.ai.ml.entities._job.pipeline._io.NodeOutput,
        ~azure.ai.ml.entities._job.pipeline._io.PipelineInput]
    :param max_concurrency: Maximum number of concurrent iterations to run. All loop body nodes will be executed
        in parallel if not specified.
    :type max_concurrency: int
    """

    OUT_TYPE_MAPPING = {
        AssetTypes.URI_FILE: AssetTypes.MLTABLE,
        AssetTypes.URI_FOLDER: AssetTypes.MLTABLE,
        AssetTypes.MLTABLE: AssetTypes.MLTABLE,
        AssetTypes.MLFLOW_MODEL: AssetTypes.MLTABLE,
        AssetTypes.TRITON_MODEL: AssetTypes.MLTABLE,
        AssetTypes.CUSTOM_MODEL: AssetTypes.MLTABLE,
        # legacy path support
        "path": AssetTypes.MLTABLE,
        ComponentParameterTypes.NUMBER: ComponentParameterTypes.STRING,
        ComponentParameterTypes.STRING: ComponentParameterTypes.STRING,
        ComponentParameterTypes.BOOLEAN: ComponentParameterTypes.STRING,
        ComponentParameterTypes.INTEGER: ComponentParameterTypes.STRING,
    }

    def __init__(
        self,
        *,
        body: "Pipeline",
        items: Union[list, dict, str, PipelineInput, NodeOutput],
        max_concurrency: Optional[int] = None,
        **kwargs: Any,
    ) -> None:
        # validate init params are valid type
        validate_attribute_type(attrs_to_check=locals(), attr_type_map=self._attr_type_map())

        kwargs.pop("type", None)
        super(ParallelFor, self).__init__(
            type=ControlFlowType.PARALLEL_FOR,
            body=body,
            **kwargs,
        )
        # loop body is incomplete in submission time, so won't validate required inputs
        self.body._validate_required_input_not_provided = False
        self._outputs: dict = {}

        actual_outputs = kwargs.get("outputs", {})
        # parallel for node shares output meta with body
        try:
            outputs = self.body._component.outputs
            # transform body outputs to aggregate types when available
            self._outputs = self._build_outputs_dict(
                outputs=actual_outputs, output_definition_dict=self._convert_output_meta(outputs)
            )
        except AttributeError:
            # when body output not available, create default output builder without meta
            self._outputs = self._build_outputs_dict(outputs=actual_outputs)

        self._items = items

        self.max_concurrency = max_concurrency

    @property
    def outputs(self) -> Dict[str, Union[str, Output]]:
        """Get the outputs of the parallel for loop.

        :return: The dictionary containing the outputs of the parallel for loop.
        :rtype: dict[str, Union[str, ~azure.ai.ml.Output]]
        """
        return self._outputs

    @property
    def items(self) -> Union[list, dict, str, PipelineInput, NodeOutput]:
        """Get the loop body's input which will bind to the loop node.

        :return: The input for the loop body.
        :rtype: typing.Union[list, dict, str, ~azure.ai.ml.entities._job.pipeline._io.NodeOutput,
            ~azure.ai.ml.entities._job.pipeline._io.PipelineInput]
        """
        return self._items

    @classmethod
    def _create_schema_for_validation(cls, context: Any) -> PathAwareSchema:
        return ParallelForSchema(context=context)

    @classmethod
    def _attr_type_map(cls) -> dict:
        return {
            **super(ParallelFor, cls)._attr_type_map(),
            "items": (dict, list, str, PipelineInput, NodeOutput),
        }

    @classmethod
    # pylint: disable-next=docstring-missing-param
    def _to_rest_item(cls, item: dict) -> dict:
        """Convert item to rest object.

        :return: The rest object
        :rtype: dict
        """
        primitive_inputs, asset_inputs = {}, {}
        # validate item
        for key, val in item.items():
            if isinstance(val, Input):
                asset_inputs[key] = val
            elif isinstance(val, (PipelineInput, NodeOutput)):
                # convert binding object to string
                primitive_inputs[key] = str(val)
            else:
                primitive_inputs[key] = val
        return {
            # asset type inputs will be converted to JobInput dict:
            # {"asset_param": {"uri": "xxx", "job_input_type": "uri_file"}}
            **cls._input_entity_to_rest_inputs(input_entity=asset_inputs),
            # primitive inputs has primitive type value like this
            # {"int_param": 1}
            **primitive_inputs,
        }

    @classmethod
    # pylint: disable-next=docstring-missing-param,docstring-missing-return,docstring-missing-rtype
    def _to_rest_items(cls, items: Union[list, dict, str, NodeOutput, PipelineInput]) -> str:
        """Convert items to rest object."""
        # validate items.
        cls._validate_items(items=items, raise_error=True, body_component=None)
        result: str = ""
        # convert items to rest object
        if isinstance(items, list):
            rest_items_list = [cls._to_rest_item(item=i) for i in items]
            result = json.dumps(rest_items_list)
        elif isinstance(items, dict):
            rest_items_dict = {k: cls._to_rest_item(item=v) for k, v in items.items()}
            result = json.dumps(rest_items_dict)
        elif isinstance(items, (NodeOutput, PipelineInput)):
            result = str(items)
        elif isinstance(items, str):
            result = items
        else:
            raise UserErrorException("Unsupported items type: {}".format(type(items)))
        return result

    def _to_rest_object(self, **kwargs: Any) -> dict:
        """Convert self to a rest object for remote call.

        :return: The rest object
        :rtype: dict
        """
        rest_node = super(ParallelFor, self)._to_rest_object(**kwargs)
        # convert items to rest object
        rest_items = self._to_rest_items(items=self.items)
        rest_node.update({"items": rest_items, "outputs": self._to_rest_outputs()})
        # TODO: Bug Item number: 2897665
        res: dict = convert_ordered_dict_to_dict(rest_node)  # type: ignore
        return res

    @classmethod
    # pylint: disable-next=docstring-missing-param,docstring-missing-return,docstring-missing-rtype
    def _from_rest_item(cls, rest_item: Any) -> Dict:
        """Convert rest item to item."""
        primitive_inputs, asset_inputs = {}, {}
        for key, val in rest_item.items():
            if isinstance(val, dict) and val.get("job_input_type"):
                asset_inputs[key] = val
            else:
                primitive_inputs[key] = val
        return {**cls._from_rest_inputs(inputs=asset_inputs), **primitive_inputs}

    @classmethod
    # pylint: disable-next=docstring-missing-param,docstring-missing-return,docstring-missing-rtype
    def _from_rest_items(cls, rest_items: str) -> Union[dict, list, str]:
        """Convert items from rest object."""
        try:
            items = json.loads(rest_items)
        except json.JSONDecodeError:
            # return original items when failed to load
            return rest_items
        if isinstance(items, list):
            return [cls._from_rest_item(rest_item=i) for i in items]
        if isinstance(items, dict):
            return {k: cls._from_rest_item(rest_item=v) for k, v in items.items()}
        return rest_items

    @classmethod
    def _from_rest_object(cls, obj: dict, pipeline_jobs: dict) -> "ParallelFor":
        # pylint: disable=protected-access
        obj = BaseNode._from_rest_object_to_init_params(obj)
        obj["items"] = cls._from_rest_items(rest_items=obj.get("items", ""))
        return cls._create_instance_from_schema_dict(pipeline_jobs=pipeline_jobs, loaded_data=obj)

    @classmethod
    def _create_instance_from_schema_dict(cls, pipeline_jobs: Dict, loaded_data: Dict, **kwargs: Any) -> "ParallelFor":
        body_name = cls._get_data_binding_expression_value(loaded_data.pop("body"), regex=r"\{\{.*\.jobs\.(.*)\}\}")

        loaded_data["body"] = cls._get_body_from_pipeline_jobs(pipeline_jobs=pipeline_jobs, body_name=body_name)
        return cls(**loaded_data, **kwargs)

    def _convert_output_meta(self, outputs: Dict[str, Union[NodeOutput, Output]]) -> Dict[str, Output]:
        """Convert output meta to aggregate types.

        :param outputs: Output meta
        :type outputs: Dict[str, Union[NodeOutput, Output]]
        :return: Dictionary of aggregate types
        :rtype: Dict[str, Output]
        """
        # pylint: disable=protected-access
        aggregate_outputs = {}
        for name, output in outputs.items():
            if output.type in self.OUT_TYPE_MAPPING:
                new_type = self.OUT_TYPE_MAPPING[output.type]
            else:
                # when loop body introduces some new output type, this will be raised as a reminder to support is in
                # parallel for
                raise UserErrorException(
                    "Referencing output with type {} is not supported in parallel_for node.".format(output.type)
                )
            if isinstance(output, NodeOutput):
                output = output._to_job_output()  # type: ignore
            if isinstance(output, Output):
                out_dict = output._to_dict()
                out_dict["type"] = new_type
                resolved_output = Output(**out_dict)
            else:
                resolved_output = Output(type=new_type)
            aggregate_outputs[name] = resolved_output
        return aggregate_outputs

    def _customized_validate(self) -> MutableValidationResult:
        """Customized validation for parallel for node.

        :return: The validation result
        :rtype: MutableValidationResult
        """
        # pylint: disable=protected-access
        validation_result = self._validate_body()
        validation_result.merge_with(
            self._validate_items(items=self.items, raise_error=False, body_component=self.body._component)
        )
        return validation_result

    @classmethod
    def _validate_items(
        cls,
        items: Union[list, dict, str, NodeOutput, PipelineInput],
        raise_error: bool = True,
        body_component: Optional[Union[str, Component]] = None,
    ) -> MutableValidationResult:
        validation_result = cls._create_empty_validation_result()
        if items is not None:
            if isinstance(items, str):
                # TODO: remove the validation
                # try to deserialize str if it's a json string
                try:
                    items = json.loads(items)
                except json.JSONDecodeError as e:
                    if not is_data_binding_expression(items, ["parent"]):
                        validation_result.append_error(
                            yaml_path="items",
                            message=f"Items is neither a valid JSON string due to {e} or a binding string.",
                        )
            if isinstance(items, dict):
                # Validate dict keys
                items = list(items.values())
            if isinstance(items, list):
                if len(items) > 0:
                    cls._validate_items_list(items, validation_result, body_component=body_component)
                else:
                    validation_result.append_error(yaml_path="items", message="Items is an empty list/dict.")
        else:
            validation_result.append_error(
                yaml_path="items",
                message="Items is required for parallel_for node",
            )
        return cls._try_raise(validation_result, raise_error=raise_error)

    @classmethod
    def _validate_items_list(
        cls,
        items: list,
        validation_result: MutableValidationResult,
        body_component: Optional[Union[str, Component]] = None,
    ) -> None:
        meta: dict = {}
        # all items have to be dict and have matched meta
        for item in items:
            # item has to be dict
            # Note: item can be empty dict when loop_body don't have foreach inputs.
            if not isinstance(item, dict):
                validation_result.append_error(
                    yaml_path="items",
                    message=f"Items has to be list/dict of dict as value, " f"but got {type(item)} for {item}.",
                )
            else:
                # item has to have matched meta
                if meta.keys() != item.keys():
                    if not meta.keys():
                        meta = item
                    else:
                        msg = f"Items should have same keys with body inputs, but got {item.keys()} and {meta.keys()}."
                        validation_result.append_error(yaml_path="items", message=msg)
                # items' keys should appear in body's inputs
                if isinstance(body_component, Component) and (not item.keys() <= body_component.inputs.keys()):
                    msg = f"Item {item} got unmatched inputs with loop body component inputs {body_component.inputs}."
                    validation_result.append_error(yaml_path="items", message=msg)
                # validate item value type
                cls._validate_item_value_type(item=item, validation_result=validation_result)

    @classmethod
    def _validate_item_value_type(cls, item: dict, validation_result: MutableValidationResult) -> None:
        supported_types = (Input, str, bool, int, float, PipelineInput)
        for _, val in item.items():
            if not isinstance(val, supported_types):
                validation_result.append_error(
                    yaml_path="items",
                    message="Unsupported type {} in parallel_for items. Supported types are: {}".format(
                        type(val), supported_types
                    ),
                )
            if isinstance(val, Input):
                cls._validate_input_item_value(entry=val, validation_result=validation_result)

    @classmethod
    def _validate_input_item_value(cls, entry: Input, validation_result: MutableValidationResult) -> None:
        if not isinstance(entry, Input):
            return
        if not entry.path:
            validation_result.append_error(
                yaml_path="items",
                message=f"Input path not provided for {entry}.",
            )
        if isinstance(entry.path, str) and os.path.exists(entry.path):
            validation_result.append_error(
                yaml_path="items",
                message=f"Local file input {entry} is not supported, please create it as a dataset.",
            )



from pathlib import Path

from azure.ai.ml import load_component
from azure.ai.ml.dsl import pipeline

upstream_component = load_component(Path(__file__).parent / "upstream_node.yml")
downstream_component = load_component(Path(__file__).parent / "downstream_node.yml")
inner_component = load_component(Path(__file__).parent / "inner_node.yml")


@pipeline
def pipeline_component_func():
    node = inner_component()
    return {"__hash__": node.outputs["__contains__"]}


@pipeline
def pipeline_func():
    upstream_node = upstream_component()
    downstream_node = downstream_component(keys=upstream_node.outputs["items"])  # noqa: F841
    pipeline_component_func()


pipeline_job = pipeline_func()
pipeline_job.settings.default_compute = "cpu-cluster"



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import pytest

from azure.ai.ml import UserIdentityConfiguration
from azure.ai.ml._restclient.v2023_04_01_preview.models import UserIdentity as RestUserIdentity
from azure.ai.ml._restclient.v2023_04_01_preview.models import ValidationMetricType
from azure.ai.ml._restclient.v2024_01_01_preview.models import (
    LearningRateScheduler,
    LogTrainingMetrics,
    LogValidationLoss,
    MLTableJobInput,
    ModelSize,
    ObjectDetectionPrimaryMetrics,
    SamplingAlgorithmType,
    StochasticOptimizer,
)
from azure.ai.ml.automl import image_object_detection
from azure.ai.ml.constants._common import AssetTypes
from azure.ai.ml.entities._inputs_outputs import Input
from azure.ai.ml.entities._job.automl import SearchSpace
from azure.ai.ml.entities._job.automl.image import ImageModelSettingsObjectDetection, ImageObjectDetectionJob
from azure.ai.ml.sweep import BanditPolicy, Choice, Uniform


@pytest.mark.automl_test
@pytest.mark.unittest
class TestAutoMLImageObjectDetection:
    @pytest.mark.parametrize("run_type", ["single", "sweep", "automode"])
    def test_image_object_detection_task(self, run_type):
        # Create AutoML Image Object Detection task
        identity = UserIdentityConfiguration()
        image_object_detection_job = image_object_detection(
            training_data=Input(type=AssetTypes.MLTABLE, path="https://foo/bar/train.csv"),
            target_column_name="label",
            primary_metric=ObjectDetectionPrimaryMetrics.MEAN_AVERAGE_PRECISION,
            validation_data=Input(type=AssetTypes.MLTABLE, path="https://foo/bar/valid.csv"),
            # Job attributes
            compute="gpu_cluster",
            name="image_object_detection_job",
            experiment_name="foo_exp",
            tags={"foo_tag": "bar"},
            identity=identity,
        )  # type: ImageObjectDetectionJob

        if run_type == "single":
            image_object_detection_job.set_limits(timeout_minutes=60)
        elif run_type == "sweep":
            image_object_detection_job.set_limits(timeout_minutes=60, max_concurrent_trials=4, max_trials=20)
        elif run_type == "automode":
            image_object_detection_job.set_limits(timeout_minutes=60, max_trials=2, max_concurrent_trials=1)

        # image_object_detection_job.training_parameters = {
        #     "checkpoint_frequency": 1,
        #     "early_stopping": True,
        #     "early_stopping_delay": 2,
        #     "early_stopping_patience": 2,
        #     "evaluation_frequency": 1,
        # }
        image_object_detection_job.set_training_parameters(
            checkpoint_frequency=1,
            early_stopping=True,
            early_stopping_delay=2,
            early_stopping_patience=2,
            evaluation_frequency=1,
        )

        if run_type == "sweep":
            """
            image_object_detection_job.search_space = [
                {
                    "model_name": 'yolov5',
                    "learning_rate": Uniform(0.0001, 0.01),
                    "model_size": Choice(['small', 'medium']),
                },
                {
                    "model_name": 'fasterrcnn_resnet50_fpn',
                    "learning_rate": Uniform(0.0001, 0.01),
                    "optimizer": Choice(['sgd', 'adam', 'adamw']),
                    "min_size": Choice([600, 800]),
                },
            ]
            """
            search_sub_space_1 = SearchSpace(
                model_name="yolov5",
                learning_rate=Uniform(0.0001, 0.01),
                model_size=Choice(["small", "medium"]),
            )
            search_sub_space_2 = SearchSpace(
                model_name="fasterrcnn_resnet50_fpn",
                learning_rate=Uniform(0.0001, 0.01),
                optimizer=Choice(["sgd", "adam", "adamw"]),
                min_size=Choice([600, 800]),
            )
            image_object_detection_job.extend_search_space([search_sub_space_1, search_sub_space_2])

            early_termination_policy = BanditPolicy(evaluation_interval=10, slack_factor=0.2)
            # image_object_detection_job.sweep = {
            #     "max_concurrent_trials": 4,
            #     "max_trials": 20,
            #     "sampling_algorithm": SamplingAlgorithmType.GRID,
            #     "early_termination": early_termination_policy,
            # }
            image_object_detection_job.set_sweep(
                sampling_algorithm=SamplingAlgorithmType.GRID,
                early_termination=early_termination_policy,
            )

        # check the rest object
        rest_obj = image_object_detection_job._to_rest_object()
        assert isinstance(rest_obj.properties.identity, RestUserIdentity)

        def _check_data_type(data, expected_type, expected_path, msg):
            if expected_type == MLTableJobInput:
                assert isinstance(data, MLTableJobInput), "{} data is not MLTableJobInput".format(msg)
            elif expected_type == Input:
                assert isinstance(data, Input), "{} data is not Input".format(msg)
                assert data.type == AssetTypes.MLTABLE, "{} data type not set correctly".format(msg)
                assert data.path == expected_path, "{} data path not set correctly".format(msg)

        _check_data_type(rest_obj.properties.task_details.training_data, MLTableJobInput, None, "Training")
        _check_data_type(rest_obj.properties.task_details.validation_data, MLTableJobInput, None, "Validation")

        original_obj = ImageObjectDetectionJob._from_rest_object(rest_obj)
        assert image_object_detection_job == original_obj, "Conversion to/from rest object failed"
        assert original_obj.compute == "gpu_cluster", "Compute not set correctly"
        assert original_obj.name == "image_object_detection_job", "Name not set correctly"
        assert original_obj.experiment_name == "foo_exp", "Experiment name not set correctly"
        assert original_obj.tags == {"foo_tag": "bar"}, "Tags not set correctly"
        assert original_obj.identity == identity
        # check if the original job inputs were restored
        _check_data_type(original_obj.training_data, Input, "https://foo/bar/train.csv", "Training")
        _check_data_type(original_obj.validation_data, Input, "https://foo/bar/valid.csv", "Validation")

    @pytest.mark.parametrize(
        "settings, expected",
        [
            (
                ("adam", "warmup_cosine", "coco_voc", "large", "enable", "enable"),
                (
                    StochasticOptimizer.ADAM,
                    LearningRateScheduler.WARMUP_COSINE,
                    ValidationMetricType.COCO_VOC,
                    ModelSize.LARGE,
                    LogTrainingMetrics.ENABLE,
                    LogValidationLoss.ENABLE,
                ),
            ),
            (
                ("Adam", "WarmupCosine", "CocoVoc", "Large", "Enable", "Enable"),
                (
                    StochasticOptimizer.ADAM,
                    LearningRateScheduler.WARMUP_COSINE,
                    ValidationMetricType.COCO_VOC,
                    ModelSize.LARGE,
                    LogTrainingMetrics.ENABLE,
                    LogValidationLoss.ENABLE,
                ),
            ),
            (
                (None, None, "coco_voc", "large", "enable", "enable"),
                (
                    None,
                    None,
                    ValidationMetricType.COCO_VOC,
                    ModelSize.LARGE,
                    LogTrainingMetrics.ENABLE,
                    LogValidationLoss.ENABLE,
                ),
            ),
        ],
        ids=["snake case", "camel case", "with None"],
    )
    def test_image_set_training_parameters_with_valid_values(self, settings, expected):
        image_object_detection_job = image_object_detection(
            training_data=Input(type=AssetTypes.MLTABLE, path="https://foo/bar/train.csv"),
            target_column_name="label",
        )
        image_object_detection_job.set_training_parameters(
            optimizer=settings[0],
            learning_rate_scheduler=settings[1],
            validation_metric_type=settings[2],
            model_size=settings[3],
            log_training_metrics=settings[4],
            log_validation_loss=settings[5],
        )
        assert image_object_detection_job.training_parameters.optimizer == expected[0]
        assert image_object_detection_job.training_parameters.learning_rate_scheduler == expected[1]
        assert image_object_detection_job.training_parameters.validation_metric_type == expected[2]
        assert image_object_detection_job.training_parameters.model_size == expected[3]
        assert image_object_detection_job.training_parameters.log_training_metrics == expected[4]
        assert image_object_detection_job.training_parameters.log_validation_loss == expected[5]

    @pytest.mark.parametrize(
        "settings, expected",
        [
            (("adamW", None, None, None, "Enable", "Enable"), pytest.raises(KeyError)),
            ((None, "Warmup_Cosine", None, None, "Enable", "Enable"), pytest.raises(KeyError)),
            ((None, None, "Coco_Voc", "large", "Enable", "Enable"), pytest.raises(KeyError)),
            ((None, None, None, "Extra_Large", "Enable", "Enable"), pytest.raises(KeyError)),
            ((None, None, None, None, "false", "Enable"), pytest.raises(KeyError)),
            ((None, None, None, None, "Enable", "false"), pytest.raises(KeyError)),
        ],
        ids=[
            "optimizer invalid",
            "learning rate scheduler invalid",
            "validation metric type invalid",
            "model size invalid",
            "log_training_metrics invalid",
            "log_validation_loss invalid",
        ],
    )
    def test_image_set_training_parameters_with_invalid_values(self, settings, expected):
        with expected:
            image_object_detection_job = image_object_detection(
                training_data=Input(type=AssetTypes.MLTABLE, path="https://foo/bar/train.csv"),
                target_column_name="label",
            )  # type: ImageClassificationJob
            image_object_detection_job.set_training_parameters(
                optimizer=settings[0],
                learning_rate_scheduler=settings[1],
                validation_metric_type=settings[2],
                model_size=settings[3],
                log_training_metrics=settings[4],
                log_validation_loss=settings[5],
            )

    @pytest.mark.parametrize(
        "settings, expected",
        [
            (
                ("adam", "warmup_cosine", "coco_voc", "large", "enable", "enable"),
                (
                    StochasticOptimizer.ADAM,
                    LearningRateScheduler.WARMUP_COSINE,
                    ValidationMetricType.COCO_VOC,
                    ModelSize.LARGE,
                    LogTrainingMetrics.ENABLE,
                    LogValidationLoss.ENABLE,
                ),
            ),
            (
                ("Adam", "WarmupCosine", "CocoVoc", "Large", "Enable", "Enable"),
                (
                    StochasticOptimizer.ADAM,
                    LearningRateScheduler.WARMUP_COSINE,
                    ValidationMetricType.COCO_VOC,
                    ModelSize.LARGE,
                    LogTrainingMetrics.ENABLE,
                    LogValidationLoss.ENABLE,
                ),
            ),
            (
                (None, None, "coco_voc", "large", "enable", "enable"),
                (
                    None,
                    None,
                    ValidationMetricType.COCO_VOC,
                    ModelSize.LARGE,
                    LogTrainingMetrics.ENABLE,
                    LogValidationLoss.ENABLE,
                ),
            ),
        ],
        ids=["snake case", "camel case", "with None"],
    )
    def test_image_set_training_parameters_with_settings_object(self, settings, expected):
        image_model_settings = ImageModelSettingsObjectDetection(
            optimizer=settings[0],
            learning_rate_scheduler=settings[1],
            validation_metric_type=settings[2],
            model_size=settings[3],
            log_training_metrics=settings[4],
            log_validation_loss=settings[5],
        )
        image_object_detection_job = image_object_detection(
            training_data=Input(type=AssetTypes.MLTABLE, path="https://foo/bar/train.csv"),
            target_column_name="label",
            training_parameters=image_model_settings,
        )
        assert image_object_detection_job.training_parameters.optimizer == expected[0]
        assert image_object_detection_job.training_parameters.learning_rate_scheduler == expected[1]
        assert image_object_detection_job.training_parameters.validation_metric_type == expected[2]
        assert image_object_detection_job.training_parameters.model_size == expected[3]
        assert image_object_detection_job.training_parameters.log_training_metrics == expected[4]
        assert image_object_detection_job.training_parameters.log_validation_loss == expected[5]



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access

from pathlib import Path
from typing import Any, Dict, Optional, Union

from azure.ai.ml._restclient.v2023_04_01_preview.models import (
    AzureDataLakeGen1Datastore as RestAzureDatalakeGen1Datastore,
)
from azure.ai.ml._restclient.v2023_04_01_preview.models import Datastore as DatastoreData
from azure.ai.ml._restclient.v2023_04_01_preview.models import DatastoreType
from azure.ai.ml._schema._datastore.adls_gen1 import AzureDataLakeGen1Schema
from azure.ai.ml.constants._common import BASE_PATH_CONTEXT_KEY, TYPE
from azure.ai.ml.entities._credentials import CertificateConfiguration, ServicePrincipalConfiguration
from azure.ai.ml.entities._datastore.datastore import Datastore
from azure.ai.ml.entities._datastore.utils import from_rest_datastore_credentials
from azure.ai.ml.entities._util import load_from_dict


class AzureDataLakeGen1Datastore(Datastore):
    """Azure Data Lake aka Gen 1 datastore that is linked to an Azure ML workspace.

    :param name: Name of the datastore.
    :type name: str
    :param store_name: Name of the Azure storage resource.
    :type store_name: str
    :param description: Description of the resource.
    :type description: str
    :param tags: Tag dictionary. Tags can be added, removed, and updated.
    :type tags: dict[str, str]
    :param properties: The asset property dictionary.
    :type properties: dict[str, str]
    :param credentials: Credentials to use for Azure ML workspace to connect to the storage.
    :type credentials: Union[ServicePrincipalSection, CertificateSection]
    :param kwargs: A dictionary of additional configuration parameters.
    :type kwargs: dict
    """

    def __init__(
        self,
        *,
        name: str,
        store_name: str,
        description: Optional[str] = None,
        tags: Optional[Dict] = None,
        properties: Optional[Dict] = None,
        credentials: Optional[Union[CertificateConfiguration, ServicePrincipalConfiguration]] = None,
        **kwargs: Any
    ):
        kwargs[TYPE] = DatastoreType.AZURE_DATA_LAKE_GEN1
        super().__init__(
            name=name, description=description, tags=tags, properties=properties, credentials=credentials, **kwargs
        )

        self.store_name = store_name

    def _to_rest_object(self) -> DatastoreData:
        gen1_ds = RestAzureDatalakeGen1Datastore(
            credentials=self.credentials._to_datastore_rest_object(),
            store_name=self.store_name,
            description=self.description,
            tags=self.tags,
        )
        return DatastoreData(properties=gen1_ds)

    @classmethod
    def _load_from_dict(
        cls, data: Dict, context: Dict, additional_message: str, **kwargs: Any
    ) -> "AzureDataLakeGen1Datastore":
        res: AzureDataLakeGen1Datastore = load_from_dict(
            AzureDataLakeGen1Schema, data, context, additional_message, **kwargs
        )
        return res

    @classmethod
    def _from_rest_object(cls, datastore_resource: DatastoreData) -> "AzureDataLakeGen1Datastore":
        properties: RestAzureDatalakeGen1Datastore = datastore_resource.properties
        return AzureDataLakeGen1Datastore(
            id=datastore_resource.id,
            name=datastore_resource.name,
            store_name=properties.store_name,
            credentials=from_rest_datastore_credentials(properties.credentials),  # type: ignore[arg-type]
            description=properties.description,
            tags=properties.tags,
        )

    def __eq__(self, other: Any) -> bool:
        res: bool = (
            super().__eq__(other)
            and self.name == other.name
            and self.type == other.type
            and self.store_name == other.store_name
            and self.credentials == other.credentials
        )
        return res

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def _to_dict(self) -> Dict:
        context = {BASE_PATH_CONTEXT_KEY: Path(".").parent}
        res: dict = AzureDataLakeGen1Schema(context=context).dump(self)
        return res



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

import logging
from typing import Any

from marshmallow import fields, post_load

from azure.ai.ml._schema.core.schema import PathAwareSchema

module_logger = logging.getLogger(__name__)


class CodeConfigurationSchema(PathAwareSchema):
    code = fields.Str()
    scoring_script = fields.Str()

    @post_load
    def make(self, data: Any, **kwargs: Any) -> Any:
        from azure.ai.ml.entities import CodeConfiguration

        return CodeConfiguration(**data)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

import logging
from typing import Any

from marshmallow import fields, post_load

from azure.ai.ml._schema.core.schema import PatchedSchemaMeta
from azure.ai.ml.entities._deployment.deployment_settings import BatchRetrySettings

module_logger = logging.getLogger(__name__)


class BatchRetrySettingsSchema(metaclass=PatchedSchemaMeta):
    max_retries = fields.Int(
        metadata={"description": "The number of maximum tries for a failed or timeout mini batch."},
    )
    timeout = fields.Int(metadata={"description": "The timeout for a mini batch."})

    @post_load
    def make(self, data: Any, **kwargs: Any) -> BatchRetrySettings:
        return BatchRetrySettings(**data)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
import typing

import strictyaml


class _SafeLoaderWithBaseLoader(strictyaml.ruamel.SafeLoader):
    """This is a SafeLoader with base resolver instead of version default resolver.

    Differences between BaseResolver and VersionedResolver:
    1) BaseResolver won't try to resolve node value. For example, "yes" and "no" will be resolved to "true"(bool)
    and "false"(bool) by VersionedResolver, but won't be resolved by BaseResolver.
    2) VersionedResolver will delay loading the pattern matching rules to pass yaml versions on loading.

    Given SafeLoader inherits from VersionedResolver, we can't directly remove VersionedResolver
    from the inheritance list. Instead, we overwrite add_version_implicit_resolver method to make
    _SafeLoaderWithBaseLoader._version_implicit_resolver empty. Then the resolver will act like a BaseResolver.
    """

    def fetch_comment(self, comment):
        pass

    def add_version_implicit_resolver(self, version, tag, regexp, first):
        """Overwrite the method to make the resolver act like a base resolver instead of version default resolver.

        :param version: version of yaml, like (1, 1)(yaml 1.1) and (1, 2)(yaml 1.2)
        :type version: VersionType
        :param tag: a tag indicating the type of the resolved node, e.g., tag:yaml.org,2002:bool.
        :type tag: Any
        :param regexp: the regular expression to match the node to be resolved
        :type regexp: Any
        :param first: a list of first characters to match
        :type first: Any
        """
        self._version_implicit_resolver.setdefault(version, {})


def yaml_safe_load_with_base_resolver(stream: typing.IO):
    """Load yaml string with base resolver instead of version default resolver.

    For example:
    1) "yes" and "no" will be loaded as "yes"(string) and "no"(string) instead of "true"(bool) and "false"(bool);
    2) "0.10" will be loaded as "0.10"(string) instead of "0.1"(float).
    3) "2019-01-01" will be loaded as "2019-01-01"(string) instead of "2019-01-01T00:00:00Z"(datetime).
    4) "1" will be loaded as "1"(string) instead of "1"(int).
    5) "1.0" will be loaded as "1.0"(string) instead of "1.0"(float).
    6) "~" will be loaded as "~"(string) instead of "None"(NoneType).

    Please refer to strictyaml.ruamel.resolver.implicit_resolvers for more details.

    :param stream: A readable stream
    :type stream: typing.IO
    :return: The return value of strictyaml.ruamel.load
    :rtype: Any
    """
    return strictyaml.ruamel.load(stream, Loader=_SafeLoaderWithBaseLoader)



# pylint: disable=too-many-lines
# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
from typing import TYPE_CHECKING

from msrest import Serializer

from azure.core.exceptions import ClientAuthenticationError, HttpResponseError, ResourceExistsError, ResourceNotFoundError, map_error
from azure.core.paging import ItemPaged
from azure.core.pipeline import PipelineResponse
from azure.core.pipeline.transport import HttpResponse
from azure.core.rest import HttpRequest
from azure.core.tracing.decorator import distributed_trace
from azure.mgmt.core.exceptions import ARMErrorFormat

from .. import models as _models
from .._vendor import _convert_request, _format_url_section

if TYPE_CHECKING:
    # pylint: disable=unused-import,ungrouped-imports
    from typing import Any, Callable, Dict, Iterable, List, Optional, TypeVar, Union
    T = TypeVar('T')
    ClsType = Optional[Callable[[PipelineResponse[HttpRequest, HttpResponse], T, Dict[str, Any]], Any]]

_SERIALIZER = Serializer()
_SERIALIZER.client_side_validation = False
# fmt: off

def build_list_by_compute_request(
    subscription_id,  # type: str
    resource_group_name,  # type: str
    workspace_name,  # type: str
    compute_name,  # type: str
    **kwargs  # type: Any
):
    # type: (...) -> HttpRequest
    filter = kwargs.pop('filter', None)  # type: Optional[str]
    continuationtoken = kwargs.pop('continuationtoken', None)  # type: Optional[str]
    orderby = kwargs.pop('orderby', None)  # type: Optional[List[str]]
    sortorder = kwargs.pop('sortorder', None)  # type: Optional[Union[str, "_models.SortOrderDirection"]]
    top = kwargs.pop('top', None)  # type: Optional[int]
    count = kwargs.pop('count', None)  # type: Optional[bool]

    accept = "application/json"
    # Construct URL
    _url = kwargs.pop("template_url", "/history/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/computes/{computeName}/runs")  # pylint: disable=line-too-long
    path_format_arguments = {
        "subscriptionId": _SERIALIZER.url("subscription_id", subscription_id, 'str'),
        "resourceGroupName": _SERIALIZER.url("resource_group_name", resource_group_name, 'str'),
        "workspaceName": _SERIALIZER.url("workspace_name", workspace_name, 'str'),
        "computeName": _SERIALIZER.url("compute_name", compute_name, 'str'),
    }

    _url = _format_url_section(_url, **path_format_arguments)

    # Construct parameters
    _query_parameters = kwargs.pop("params", {})  # type: Dict[str, Any]
    if filter is not None:
        _query_parameters['$filter'] = _SERIALIZER.query("filter", filter, 'str')
    if continuationtoken is not None:
        _query_parameters['$continuationtoken'] = _SERIALIZER.query("continuationtoken", continuationtoken, 'str')
    if orderby is not None:
        _query_parameters['$orderby'] = _SERIALIZER.query("orderby", orderby, '[str]')
    if sortorder is not None:
        _query_parameters['$sortorder'] = _SERIALIZER.query("sortorder", sortorder, 'str')
    if top is not None:
        _query_parameters['$top'] = _SERIALIZER.query("top", top, 'int')
    if count is not None:
        _query_parameters['$count'] = _SERIALIZER.query("count", count, 'bool')

    # Construct headers
    _header_parameters = kwargs.pop("headers", {})  # type: Dict[str, Any]
    _header_parameters['Accept'] = _SERIALIZER.header("accept", accept, 'str')

    return HttpRequest(
        method="GET",
        url=_url,
        params=_query_parameters,
        headers=_header_parameters,
        **kwargs
    )

# fmt: on
class RunOperations(object):
    """RunOperations operations.

    You should not instantiate this class directly. Instead, you should create a Client instance that
    instantiates it for you and attaches it as an attribute.

    :ivar models: Alias to model classes used in this operation group.
    :type models: ~azure.mgmt.machinelearningservices.models
    :param client: Client for service requests.
    :param config: Configuration of service client.
    :param serializer: An object model serializer.
    :param deserializer: An object model deserializer.
    """

    models = _models

    def __init__(self, client, config, serializer, deserializer):
        self._client = client
        self._serialize = serializer
        self._deserialize = deserializer
        self._config = config

    @distributed_trace
    def list_by_compute(
        self,
        subscription_id,  # type: str
        resource_group_name,  # type: str
        workspace_name,  # type: str
        compute_name,  # type: str
        filter=None,  # type: Optional[str]
        continuationtoken=None,  # type: Optional[str]
        orderby=None,  # type: Optional[List[str]]
        sortorder=None,  # type: Optional[Union[str, "_models.SortOrderDirection"]]
        top=None,  # type: Optional[int]
        count=None,  # type: Optional[bool]
        **kwargs  # type: Any
    ):
        # type: (...) -> Iterable["_models.PaginatedRunList"]
        """list_by_compute.

        :param subscription_id: The Azure Subscription ID.
        :type subscription_id: str
        :param resource_group_name: The Name of the resource group in which the workspace is located.
        :type resource_group_name: str
        :param workspace_name: The name of the workspace.
        :type workspace_name: str
        :param compute_name:
        :type compute_name: str
        :param filter: Allows for filtering the collection of resources.
         The expression specified is evaluated for each resource in the collection, and only items
         where the expression evaluates to true are included in the response.
        :type filter: str
        :param continuationtoken: The continuation token to use for getting the next set of resources.
        :type continuationtoken: str
        :param orderby: The list of resource properties to use for sorting the requested resources.
        :type orderby: list[str]
        :param sortorder: The sort order of the returned resources. Not used, specify asc or desc after
         each property name in the OrderBy parameter.
        :type sortorder: str or ~azure.mgmt.machinelearningservices.models.SortOrderDirection
        :param top: The maximum number of items in the resource collection to be included in the
         result.
         If not specified, all items are returned.
        :type top: int
        :param count: Whether to include a count of the matching resources along with the resources
         returned in the response.
        :type count: bool
        :keyword callable cls: A custom type or function that will be passed the direct response
        :return: An iterator like instance of either PaginatedRunList or the result of cls(response)
        :rtype:
         ~azure.core.paging.ItemPaged[~azure.mgmt.machinelearningservices.models.PaginatedRunList]
        :raises: ~azure.core.exceptions.HttpResponseError
        """
        cls = kwargs.pop('cls', None)  # type: ClsType["_models.PaginatedRunList"]
        error_map = {
            401: ClientAuthenticationError, 404: ResourceNotFoundError, 409: ResourceExistsError
        }
        error_map.update(kwargs.pop('error_map', {}))
        def prepare_request(next_link=None):
            if not next_link:
                
                request = build_list_by_compute_request(
                    subscription_id=subscription_id,
                    resource_group_name=resource_group_name,
                    workspace_name=workspace_name,
                    compute_name=compute_name,
                    filter=filter,
                    continuationtoken=continuationtoken,
                    orderby=orderby,
                    sortorder=sortorder,
                    top=top,
                    count=count,
                    template_url=self.list_by_compute.metadata['url'],
                )
                request = _convert_request(request)
                request.url = self._client.format_url(request.url)

            else:
                
                request = build_list_by_compute_request(
                    subscription_id=subscription_id,
                    resource_group_name=resource_group_name,
                    workspace_name=workspace_name,
                    compute_name=compute_name,
                    filter=filter,
                    continuationtoken=continuationtoken,
                    orderby=orderby,
                    sortorder=sortorder,
                    top=top,
                    count=count,
                    template_url=next_link,
                )
                request = _convert_request(request)
                request.url = self._client.format_url(request.url)
                request.method = "GET"
            return request

        def extract_data(pipeline_response):
            deserialized = self._deserialize("PaginatedRunList", pipeline_response)
            list_of_elem = deserialized.value
            if cls:
                list_of_elem = cls(list_of_elem)
            return deserialized.next_link or None, iter(list_of_elem)

        def get_next(next_link=None):
            request = prepare_request(next_link)

            pipeline_response = self._client._pipeline.run(  # pylint: disable=protected-access
                request,
                stream=False,
                **kwargs
            )
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)
                error = self._deserialize.failsafe_deserialize(_models.ErrorResponse, pipeline_response)
                raise HttpResponseError(response=response, model=error, error_format=ARMErrorFormat)

            return pipeline_response


        return ItemPaged(
            get_next, extract_data
        )
    list_by_compute.metadata = {'url': "/history/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/computes/{computeName}/runs"}  # type: ignore



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
from typing import Dict, Optional, Union

from azure.ai.ml.entities._builders import Command
from azure.ai.ml.entities._builders.do_while import DoWhile
from azure.ai.ml.entities._builders.pipeline import Pipeline
from azure.ai.ml.entities._inputs_outputs import Output
from azure.ai.ml.entities._job.pipeline._io import NodeOutput


def do_while(
    body: Union[Pipeline, Command], mapping: Dict, max_iteration_count: int, condition: Optional[Output] = None
) -> DoWhile:
    """Build a do_while node by specifying the loop body, output-input mapping, and termination condition.

    .. note::
        The following example shows how to use the `do_while` function to create a pipeline with a `do_while` node.

        .. code-block:: python

            from azure.ai.ml.dsl import pipeline
            from mldesigner.dsl import do_while

            @pipeline()
            def your_do_while_body():
                pass

            @pipeline()
            def pipeline_with_do_while_node():
                do_while_body = your_do_while_body()
                do_while_node = do_while(
                    body=do_while_body,
                    condition=do_while_body.outputs.condition_output,
                    mapping={
                        do_while_body.outputs.output1: do_while_body_inputs.input1,
                        do_while_body.outputs.output2: [
                            do_while_body_inputs.input2,
                            do_while_body_inputs.input3,
                        ],
                    },
                )
                # Connect to the do_while_node outputs
                component = component_func(
                    input1=do_while_body.outputs.output1, input2=do_while_body.outputs.output2
                )

    :param body: The pipeline job or command node for the do-while loop body.
    :type body: Union[~azure.ai.ml.entities._builders.pipeline.Pipeline, ~azure.ai.ml.entities._builders.Command]
    :param mapping: The output-input mapping for each round of the do-while loop.
        The key is the last round's output of the body, and the value is the input port for the current body.
    :type mapping: Dict[
        Union[str,  ~azure.ai.ml.entities.Output],
        Union[str, ~azure.ai.ml.entities.Input, List]]
    :param max_iteration_count: The limit on running the do-while node.
    :type max_iteration_count: int
    :param condition: The name of a boolean output of the body.
        The do-while loop stops if its value is evaluated to be negative.
        If not specified, it handles as a while-true loop.
    :type condition:  ~azure.ai.ml.entities.Output
    :return: The do-while node.
    :rtype: ~azure.ai.ml.entities._builders.do_while.DoWhile
    """
    do_while_node = DoWhile(
        body=body,
        condition=condition,  # type: ignore[arg-type]
        mapping=mapping,
        _from_component_func=True,
    )
    do_while_node.set_limits(max_iteration_count=max_iteration_count)

    def _infer_and_update_body_input_from_mapping() -> None:
        # pylint: disable=protected-access
        for source_output, body_input in mapping.items():
            # handle case that mapping key is a NodeOutput
            output_name = source_output._port_name if isinstance(source_output, NodeOutput) else source_output
            # if loop body output type is not specified, skip as we have no place to infer
            if body.outputs[output_name].type is None:
                continue
            # body input can be a list of inputs, normalize as a list to process
            if not isinstance(body_input, list):
                body_input = [body_input]
            for single_input in body_input:
                # if input type is specified, no need to infer and skip
                if single_input.type is not None:
                    continue
                inferred_type = body.outputs[output_name].type
                # update node input
                single_input._meta._is_inferred_optional = True
                single_input.type = inferred_type
                # update node corresponding component input
                input_name = single_input._meta.name
                body.component.inputs[input_name]._is_inferred_optional = True  # type: ignore[union-attr]
                body.component.inputs[input_name].type = inferred_type  # type: ignore[union-attr]

    # when mapping is a dictionary, infer and update for dynamic input
    if isinstance(mapping, dict):
        _infer_and_update_body_input_from_mapping()

    return do_while_node
